def rename_into_caps(
    in_bids_dwi,
    fname_dwi,
    fname_bval,
    fname_bvec,
    fname_brainmask,
    fname_magnitude,
    fname_fmap,
    fname_smoothed_fmap,
):
    """Rename the outputs of the pipelines into CAPS.

    Args:
        in_bids_dwi (str): Input BIDS DWI to extract the <source_file>
        fname_dwi (str): Preprocessed DWI file.
        fname_bval (str): Preprocessed bval.
        fname_bvec (str): Preprocessed bvec.
        fname_brainmask (str): B0 mask.
        fname_smoothed_fmap (str): Smoothed (calibrated) fmap on b0 space.
        fname_fmap (str): Calibrated fmap on b0 space.
        fname_magnitude (str): Magnitude image on b0 space.

    Returns:
        Tuple[str, str, str, str, str, str, str]: The different outputs in CAPS format.
    """
    import os

    from nipype.interfaces.utility import Rename
    from nipype.utils.filemanip import split_filename

    # Extract <source_file> in format sub-CLNC01_ses-M00_[acq-label]_dwi
    _, source_file_dwi, _ = split_filename(in_bids_dwi)

    # Extract base path from fname:
    base_dir_dwi, _, _ = split_filename(fname_dwi)
    base_dir_bval, _, _ = split_filename(fname_bval)
    base_dir_bvec, _, _ = split_filename(fname_bvec)
    base_dir_brainmask, _, _ = split_filename(fname_brainmask)
    base_dir_smoothed_fmap, _, _ = split_filename(fname_smoothed_fmap)
    base_dir_calibrated_fmap, _, _ = split_filename(fname_fmap)
    base_dir_magnitude, _, _ = split_filename(fname_magnitude)

    # Rename into CAPS DWI:
    rename_dwi = Rename()
    rename_dwi.inputs.in_file = fname_dwi
    rename_dwi.inputs.format_string = os.path.join(
        base_dir_dwi, f"{source_file_dwi}_space-b0_preproc.nii.gz"
    )
    out_caps_dwi = rename_dwi.run()

    # Rename into CAPS bval:
    rename_bval = Rename()
    rename_bval.inputs.in_file = fname_bval
    rename_bval.inputs.format_string = os.path.join(
        base_dir_bval, f"{source_file_dwi}_space-b0_preproc.bval"
    )
    out_caps_bval = rename_bval.run()

    # Rename into CAPS bvec:
    rename_bvec = Rename()
    rename_bvec.inputs.in_file = fname_bvec
    rename_bvec.inputs.format_string = os.path.join(
        base_dir_bvec, f"{source_file_dwi}_space-b0_preproc.bvec"
    )
    out_caps_bvec = rename_bvec.run()

    # Rename into CAPS brainmask:
    rename_brainmask = Rename()
    rename_brainmask.inputs.in_file = fname_brainmask
    rename_brainmask.inputs.format_string = os.path.join(
        base_dir_brainmask, f"{source_file_dwi}_space-b0_brainmask.nii.gz"
    )
    out_caps_brainmask = rename_brainmask.run()

    # Rename into CAPS magnitude:
    rename_magnitude = Rename()
    rename_magnitude.inputs.in_file = fname_magnitude
    rename_magnitude.inputs.format_string = os.path.join(
        base_dir_magnitude, f"{source_file_dwi}_space-b0_magnitude1.nii.gz"
    )
    out_caps_magnitude = rename_magnitude.run()

    # Rename into CAPS fmap:
    rename_calibrated_fmap = Rename()
    rename_calibrated_fmap.inputs.in_file = fname_fmap
    rename_calibrated_fmap.inputs.format_string = os.path.join(
        base_dir_calibrated_fmap, f"{source_file_dwi}_space-b0_fmap.nii.gz"
    )
    out_caps_fmap = rename_calibrated_fmap.run()

    # Rename into CAPS smoothed fmap:
    rename_smoothed_fmap = Rename()
    rename_smoothed_fmap.inputs.in_file = fname_smoothed_fmap
    rename_smoothed_fmap.inputs.format_string = os.path.join(
        base_dir_smoothed_fmap, f"{source_file_dwi}_space-b0_fwhm-4_fmap.nii.gz"
    )
    out_caps_smoothed_fmap = rename_smoothed_fmap.run()

    return (
        out_caps_dwi.outputs.out_file,
        out_caps_bval.outputs.out_file,
        out_caps_bvec.outputs.out_file,
        out_caps_brainmask.outputs.out_file,
        out_caps_magnitude.outputs.out_file,
        out_caps_fmap.outputs.out_file,
        out_caps_smoothed_fmap.outputs.out_file,
    )


def get_grad_fsl(bvec, bval):
    grad_fsl = (bvec, bval)
    return grad_fsl


def init_input_node(
    dwi, bvec, bval, dwi_json, fmap_magnitude, fmap_phasediff, fmap_phasediff_json
):
    """Initialize pipeline (read JSON, check files and print begin message)."""
    import datetime

    import nibabel as nib

    from clinica.utils.dwi import bids_dir_to_fsl_dir, check_dwi_volume
    from clinica.utils.filemanip import extract_metadata_from_json, get_subject_id
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    # Extract image ID
    image_id = get_subject_id(dwi)

    # Check that the number of DWI, bvec & bval are the same
    try:
        check_dwi_volume(dwi, bvec, bval)
    except ValueError as e:
        now = datetime.datetime.now().strftime("%H:%M:%S")
        error_msg = f"[{now}] Error: Number of DWIs, b-vals and b-vecs mismatch for {image_id.replace('_', ' | ')}"
        cprint(error_msg, lvl="error")
        raise ValueError(e)

    # Check that PhaseDiff and magnitude1 have the same header
    # Otherwise, FSL in FugueExtrapolationFromMask will crash
    img_phasediff = nib.load(fmap_phasediff)
    img_magnitude = nib.load(fmap_magnitude)
    if img_phasediff.shape != img_magnitude.shape:
        now = datetime.datetime.now().strftime("%H:%M:%S")
        error_msg = (
            f"[{now}] Error: Headers of PhaseDiff and Magnitude1 are not the same "
            f"for {image_id.replace('_', ' | ')} ({img_phasediff.shape} vs {img_magnitude.shape})"
        )
        cprint(error_msg, lvl="error")
        raise NotImplementedError(error_msg)

    # Read metadata from DWI JSON file:
    [total_readout_time, phase_encoding_direction] = extract_metadata_from_json(
        dwi_json, ["TotalReadoutTime", "PhaseEncodingDirection"]
    )
    phase_encoding_direction = bids_dir_to_fsl_dir(phase_encoding_direction)

    # Read metadata from PhaseDiff JSON file:
    [echo_time1, echo_time2] = extract_metadata_from_json(
        fmap_phasediff_json, ["EchoTime1", "EchoTime2"]
    )
    delta_echo_time = abs(echo_time2 - echo_time1)

    # Print begin message
    print_begin_image(
        image_id,
        ["TotalReadoutTime", "PhaseEncodingDirection", "DeltaEchoTime"],
        [str(total_readout_time), phase_encoding_direction, str(delta_echo_time)],
    )

    return (
        image_id,
        dwi,
        bvec,
        bval,
        total_readout_time,
        phase_encoding_direction,
        fmap_magnitude,
        fmap_phasediff,
        delta_echo_time,
    )


def print_end_pipeline(image_id, final_file):
    """Display end message for `image_id` when `final_file` is connected."""
    from clinica.utils.ux import print_end_image

    print_end_image(image_id)


def rads2hz(in_file, delta_te, out_file=None):
    """Convert input phase difference map to Hz."""
    import math
    import os

    import nibabel as nb
    import numpy as np

    if out_file is None:
        fname, fext = os.path.splitext(os.path.basename(in_file))
        if fext == ".gz":
            fname, _ = os.path.splitext(fname)
        out_file = os.path.abspath(f"./{fname}_radsec.nii.gz")

    im = nb.load(in_file)
    data = im.get_data().astype(np.float32) * (1.0 / (delta_te * 2 * math.pi))
    nb.Nifti1Image(data, im.affine, im.header).to_filename(out_file)
    return out_file


def resample_fmap_to_b0(in_fmap, in_b0, out_file=None):
    """Resample fieldmap onto the b0 image.

    Warnings:
        The fieldmap should already be aligned on the b0.

    Args:
        in_fmap(str): Fieldmap image.
        in_b0(str): B0 image.
        out_file(optional[str]): Filename (default: <in_fmap>_space-b0.nii.gz

    Returns:
        Fieldmap image resampled on b0.
    """
    import os

    import nibabel
    from nilearn.image import resample_to_img

    if out_file is None:
        fname, ext = os.path.splitext(os.path.basename(in_fmap))
        if ext == ".gz":
            fname, ext2 = os.path.splitext(fname)
            ext = ext2 + ext
        out_resampled_fmap = os.path.abspath(f"{fname}_space-b0{ext}")
    else:
        out_resampled_fmap = out_file

    resampled_fmap = resample_to_img(
        source_img=in_fmap, target_img=in_b0, interpolation="continuous"
    )

    nibabel.nifti1.save(resampled_fmap, out_resampled_fmap)

    return out_resampled_fmap



"""This module contains the Pipeline abstract class needed for Clinica.

Subclasses are located in clinica/pipeline/<pipeline_name>/<pipeline_name>_pipeline.py
"""

import abc
from typing import Tuple

import click
from nipype.pipeline.engine import Workflow


def postset(attribute, value):
    """Sets the attribute of an object after the execution.

    Args:
        attribute: An object's attribute to be set.
        value: A desired value for the object's attribute.

    Returns:
        A decorator executed after the decorated function is.
    """

    def postset_decorator(func):
        def func_wrapper(self, *args, **kwargs):
            res = func(self, *args, **kwargs)
            setattr(self, attribute, value)
            return res

        return func_wrapper

    return postset_decorator


def detect_cross_sectional_and_longitudinal_subjects(
    all_subs: list, bids_dir: str
) -> Tuple[list, list]:
    """This function detects whether a subject has a longitudinal or cross sectional structure.
    If it has a mixed structure, it will be considered cross sectionnal.

    Parameters
    ----------
    all_subs: list
        List containing all the names of the subjects.

    bids_dir: str
        Path to the bids directory.

    Returns
    -------
    cross_subj: list
        List of all the subject detected with a cross_sectional organisation.

    long_subj: list
        List of all the subject detected with a longitudinal organisation.

    Examples
    --------
    BIDS
    ├── sub-01
    │   └── anat
    └── sub-02
        └── ses-M000
    >>> detect_cross_sectional_and_longitudinal_subjects(["sub-01", "sub-02"], /Users/name.surname/BIDS)
    (["sub-01"], ["sub-02])
    """
    from os import listdir
    from os.path import isdir, join

    cross_subj = []
    long_subj = []
    for sub in all_subs:
        folder_list = [
            f for f in listdir(join(bids_dir, sub)) if isdir(join(bids_dir, sub, f))
        ]

        if not all([fold.startswith("ses-") for fold in folder_list]):
            cross_subj.append(sub)
        else:
            long_subj.append(sub)
    return cross_subj, long_subj


class Pipeline(Workflow):
    """Clinica Pipeline class.

    This class overwrites the `Workflow` to integrate and encourage the
    use of BIDS and CAPS data structures as inputs and outputs of the pipelines
    developed for the Clinica software.

    The global architecture of a Clinica pipelines is as follow:
        [ Data Input Stream ]
                |
            [ Input ]
                |
            [[ Core ]] <- Could be one or more `npe.Node`s
                |
            [ Output ]
                |
        [ Data Output Stream ]

    Attributes:
        is_built (bool): Informs if the pipelines has been built or not.
        parameters (dict): Parameters of the pipelines.
        info (dict): Information presented in the associated `info.json` file.
        input_node (:obj:`npe.Node`): Identity interface connecting inputs.
        output_node (:obj:`npe.Node`): Identity interface connecting outputs.
        bids_directory (str): Directory used to read the data from, in BIDS.
        caps_directory (str): Directory used to read/write the data from/to,
            in CAPS.
        subjects (list): List of subjects defined in the `subjects.tsv` file.
            # TODO(@jguillon): Check the subjects-sessions file name.
        sessions (list): List of sessions defined in the `subjects.tsv` file.
        tsv_file (str): Path to the subjects-sessions `.tsv` file.
        info_file (str): Path to the associated `info.json` file.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        bids_directory=None,
        caps_directory=None,
        tsv_file=None,
        overwrite_caps=False,
        base_dir=None,
        parameters={},
        name=None,
    ):
        """Init a Pipeline object.

        Args:
            bids_directory (str, optional): Path to a BIDS directory. Defaults to None.
            caps_directory (str, optional): Path to a CAPS directory. Defaults to None.
            tsv_file (str, optional): Path to a subjects-sessions `.tsv` file. Defaults to None.
            overwrite_caps (bool, optional): Overwrite or not output directory.. Defaults to False.
            base_dir (str, optional): Working directory (attribute of Nipype::Workflow class). Defaults to None.
            parameters (dict, optional): Pipeline parameters. Defaults to {}.
            name (str, optional): Pipeline name. Defaults to None.

        Raises:
            RuntimeError: [description]
        """
        import inspect
        import os
        from tempfile import mkdtemp

        from clinica.utils.inputs import check_bids_folder, check_caps_folder
        from clinica.utils.participant import get_subject_session_list

        self._is_built = False
        self._overwrite_caps = overwrite_caps
        self._bids_directory = bids_directory
        self._caps_directory = caps_directory
        self._verbosity = "debug"
        self._tsv_file = tsv_file
        self._info_file = os.path.join(
            os.path.dirname(os.path.abspath(inspect.getfile(self.__class__))),
            "info.json",
        )
        self._info = {}

        if base_dir:
            self.base_dir = base_dir
            self._base_dir_was_specified = True
        else:
            self.base_dir = mkdtemp()
            self._base_dir_was_specified = False

        self._name = name or self.__class__.__name__
        self._parameters = parameters

        if not self._bids_directory:
            if not self._caps_directory:
                raise RuntimeError(
                    f"The {self._name} pipeline does not contain "
                    "BIDS nor CAPS directory at the initialization."
                )

            check_caps_folder(self._caps_directory)
            input_dir = self._caps_directory
            is_bids_dir = False
        else:
            check_bids_folder(self._bids_directory)
            input_dir = self._bids_directory
            is_bids_dir = True
        self._sessions, self._subjects = get_subject_session_list(
            input_dir, tsv_file, is_bids_dir, False, base_dir
        )

        self.init_nodes()

    @staticmethod
    def get_processed_images(caps_directory, subjects, sessions):
        """Extract processed image IDs in `caps_directory` based on `subjects`_`sessions`.

        Todo:
            [ ] Implement this static method in all pipelines
            [ ] Make it abstract to force overload in future pipelines
        """
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.stream import cprint

        cprint(msg="Pipeline finished with errors.", lvl="error")
        cprint(msg="CAPS outputs were not found for some image(s):", lvl="error")
        raise ClinicaException(
            "Implementation on which image(s) failed will appear soon."
        )

    def init_nodes(self):
        """Init the basic workflow and I/O nodes necessary before build."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        if self.get_input_fields():
            self._input_node = npe.Node(
                name="Input",
                interface=nutil.IdentityInterface(
                    fields=self.get_input_fields(), mandatory_inputs=False
                ),
            )
        else:
            self._input_node = None

        if self.get_output_fields():
            self._output_node = npe.Node(
                name="Output",
                interface=nutil.IdentityInterface(
                    fields=self.get_output_fields(), mandatory_inputs=False
                ),
            )
        else:
            self._output_node = None

        Workflow.__init__(self, self._name, self.base_dir)
        if self.input_node:
            self.add_nodes([self.input_node])
        if self.output_node:
            self.add_nodes([self.output_node])

    def has_input_connections(self):
        """Checks if the Pipeline's input node has been connected.

        Returns:
            True if the input node is connected, False otherwise.
        """
        if self.input_node:
            return self._graph.in_degree(self.input_node) > 0
        else:
            return False

    def has_output_connections(self):
        """Checks if the Pipeline's output node has been connected.

        Returns:
            True if the output node is connected, False otherwise.
        """
        if self.output_node:
            return self._graph.out_degree(self.output_node) > 0
        else:
            return False

    @postset("is_built", True)
    def build(self):
        """Builds the core, input and output nodes of the Pipeline.

        This method first checks it has already been run. It then checks
        the pipelines dependencies and, in this order, builds the core nodes,
        the input node and, finally, the output node of the Pipeline.

        Since this method returns the concerned object, it can be chained to
        any other method of the Pipeline class.

        Returns:
            self: A Pipeline object.
        """
        if not self.is_built:
            self.check_dependencies()
            self.check_pipeline_parameters()
            if not self.has_input_connections():
                self.build_input_node()
            self.build_core_nodes()
            if not self.has_output_connections():
                self.build_output_node()
        return self

    def run(self, plugin=None, plugin_args=None, update_hash=False, bypass_check=False):
        """Executes the Pipeline.

        It overwrites the default Workflow method to check if the
        Pipeline is built before running it. If not, it builds it and then
        run it.
        It also checks whether there is enough space left on the disks, and if
        the number of threads to run in parallel is consistent with what is
        possible on the CPU.

        Args:
            Similar to those of Workflow.run.

        Returns:
            An execution graph (see Workflow.run).
        """
        import shutil

        from networkx import Graph, NetworkXError

        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_failed_images

        if not self.is_built:
            self.build()
        self.check_not_cross_sectional()

        if not bypass_check:
            self.check_size()
            plugin_args = self.update_parallelize_info(plugin_args)
            plugin = "MultiProc"
        exec_graph = []
        try:
            exec_graph = Workflow.run(self, plugin, plugin_args, update_hash)
            if not self.base_dir_was_specified:
                shutil.rmtree(self.base_dir)

        except RuntimeError as e:
            # Check that it is a Nipype error
            if "Workflow did not execute cleanly. Check log for details" in str(e):
                input_ids = [
                    p_id + "_" + s_id
                    for p_id, s_id in zip(self.subjects, self.sessions)
                ]
                output_ids = self.get_processed_images(
                    caps_directory=self.caps_directory,
                    subjects=self.subjects,
                    sessions=self.sessions,
                )
                missing_ids = list(set(input_ids) - set(output_ids))
                print_failed_images(self.name, missing_ids)
            else:
                raise e
        except NetworkXError:
            cprint(
                msg=(
                    "Either all the images were already run by the pipeline "
                    "or no image was found to run the pipeline."
                ),
                lvl="warning",
            )
            exec_graph = Graph()
        return exec_graph

    def load_info(self):
        """Loads the associated info.json file.

        Todo:
            - [ ] Raise an appropriate exception when the info file can't open

        Raises:
            None. # TODO(@jguillon)

        Returns:
            self: A Pipeline object.
        """
        import json

        with open(self.info_file) as info_file:
            self.info = json.load(info_file)
        return self

    def check_dependencies(self):
        """Checks if listed dependencies are present.

        Loads the pipelines related `info.json` file and check each one of the
        dependencies listed in the JSON "dependencies" field. Its raises
        exception if a program in the list does not exist or if environment
        variables are not properly defined.

        Todo:
            - [ ] MATLAB toolbox dependency checking
            - [x] check MATLAB
            - [ ] Clinica pipelines dependency checks
            - [ ] Check dependencies version

        Raises:
            Exception: Raises an exception when bad dependency types given in
            the `info.json` file are detected.

        Returns:
            self: A Pipeline object.
        """
        import clinica.utils.check_dependency as chk

        # Checking functions preparation
        check_software = {
            # 'matlab': chk.check_matlab,
            "ants": chk.check_ants,
            "convert3d": chk.check_convert3d,
            "spm": chk.check_spm,
            "freesurfer": chk.check_freesurfer,
            "fsl": chk.check_fsl,
            "mrtrix": chk.check_mrtrix,
            "matlab": chk.check_matlab,
            "petpvc": chk.check_petpvc,
        }
        check_binary = chk.is_binary_present
        # check_toolbox = chk.is_toolbox_present
        # check_pipeline = chk.is_pipeline_present

        # Load the info.json file
        if not self.info:
            self.load_info()

        # Dependencies checking
        for d in self.info["dependencies"]:
            if d["type"] == "software":
                check_software[d["name"]](d["version"])
            elif d["type"] == "binary":
                check_binary(d["name"])
            elif d["type"] == "toolbox":
                pass
            elif d["type"] == "pipeline":
                pass
            else:
                raise Exception(
                    f"Pipeline.check_dependencies() Unknown dependency type: '{d['type']}'."
                )

        self.check_custom_dependencies()

        return self

    def check_size(self):
        """Check if the pipeline has enough space on the disk for both working directory and CAPS."""
        import select
        import sys
        from os import statvfs
        from os.path import dirname

        from clinica.utils.stream import cprint

        SYMBOLS = {
            "customary": ("B", "K", "M", "G", "T", "P", "E", "Z", "Y"),
            "customary_ext": (
                "byte",
                "kilo",
                "mega",
                "giga",
                "tera",
                "peta",
                "exa",
                "zetta",
                "iotta",
            ),
            "iec": ("Bi", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"),
            "iec_ext": (
                "byte",
                "kibi",
                "mebi",
                "gibi",
                "tebi",
                "pebi",
                "exbi",
                "zebi",
                "yobi",
            ),
        }

        def bytes2human(n, format="%(value).1f %(symbol)s", symbols="customary"):
            """
            Convert n bytes into a human readable string based on format.
            Symbols can be either "customary", "customary_ext", "iec" or "iec_ext",
            see: http://goo.gl/kTQMs

            License:
            Bytes-to-human / human-to-bytes converter.
            Based on: http://goo.gl/kTQMs
            Working with Python 2.x and 3.x.
            Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
            License: MIT
            """
            n = int(n)
            if n < 0:
                raise ValueError("n < 0")
            symbols = SYMBOLS[symbols]
            prefix = {}
            for i, s in enumerate(symbols[1:]):
                prefix[s] = 1 << (i + 1) * 10
            for symbol in reversed(symbols[1:]):
                if n >= prefix[symbol]:
                    value = float(n) / prefix[symbol]
                    return format % locals()
            return format % dict(symbol=symbols[0], value=n)

        def human2bytes(s):
            """
            Attempts to guess the string format based on default symbols
            set and return the corresponding bytes as an integer.
            When unable to recognize the format ValueError is raised.

            License:
            Bytes-to-human / human-to-bytes converter.
            Based on: http://goo.gl/kTQMs
            Working with Python 2.x and 3.x.
            Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
            License: MIT
            """
            init = s
            num = ""
            while s and s[0:1].isdigit() or s[0:1] == ".":
                num += s[0]
                s = s[1:]
            num = float(num)
            letter = s.strip()
            for name, sset in SYMBOLS.items():
                if letter in sset:
                    break
            else:
                if letter == "k":
                    # treat 'k' as an alias for 'K' as per: http://goo.gl/kTQMs
                    sset = SYMBOLS["customary"]
                    letter = letter.upper()
                else:
                    raise ValueError("can't interpret %r" % init)
            prefix = {sset[0]: 1}
            for i, s in enumerate(sset[1:]):
                prefix[s] = 1 << (i + 1) * 10
            return int(num * prefix[letter])

        timeout = 15

        # Get the number of sessions
        n_sessions = len(self.subjects)
        try:
            caps_stat = statvfs(self.caps_directory)
        except FileNotFoundError:
            # CAPS folder may not exist yet
            caps_stat = statvfs(dirname(self.caps_directory))
        wd_stat = statvfs(dirname(self.base_dir))

        # Estimate space left on partition/disk/whatever caps and wd is located
        free_space_caps = caps_stat.f_bavail * caps_stat.f_frsize
        free_space_wd = wd_stat.f_bavail * wd_stat.f_frsize

        try:
            space_needed_caps_1_session = self.info["space_caps"]
            space_needed_wd_1_session = self.info["space_wd"]
            space_needed_caps = n_sessions * human2bytes(space_needed_caps_1_session)
            space_needed_wd = n_sessions * human2bytes(space_needed_wd_1_session)
            error = ""
            if free_space_caps == free_space_wd:
                if space_needed_caps + space_needed_wd > free_space_wd:
                    # We assume this is the same disk
                    error = (
                        f"{error}Space needed for CAPS and working directory ("
                        f"{bytes2human(space_needed_caps + space_needed_wd)}"
                        f") is greater than what is left on your hard drive ("
                        f"{bytes2human(free_space_wd)})"
                    )
            else:
                if space_needed_caps > free_space_caps:
                    error = (
                        f"{error} Space needed for CAPS ("
                        f"{bytes2human(space_needed_caps)}"
                        f") is greater than what is left on your hard drive ("
                        f"{bytes2human(free_space_caps)})\n"
                    )
                if space_needed_wd > free_space_wd:
                    error = (
                        f"{error}Space needed for working_directory ("
                        f"{bytes2human(space_needed_wd)}"
                        f") is greater than what is left on your hard drive ("
                        f"{bytes2human(free_space_wd)})\n"
                    )
            if error:
                cprint(msg=f"[SpaceError] {error}", lvl="error")
                if not click.confirm(
                    "Space issues detected. Do you still want to run the pipeline?"
                ):
                    click.echo("Clinica will now exit...")
                    sys.exit()

        except KeyError:
            cprint("No info on how much size the pipeline takes. Running anyway...")

    def update_parallelize_info(self, plugin_args):
        """Performs some checks of the number of threads given in parameters,
        given the number of CPUs of the machine in which clinica is running.
        We force the use of plugin MultiProc

        Author: Arnaud Marcoux"""
        import select
        import sys
        from multiprocessing import cpu_count

        from clinica.utils.stream import cprint

        # count number of CPUs
        n_cpu = cpu_count()
        # timeout value: max time allowed to decide how many thread
        # to run in parallel (sec)
        timeout = 15

        # Use this var to know in the end if we need to ask the user
        # an other number
        ask_user = False

        try:
            # if no --n_procs arg is used, plugin_arg is None
            # so we need a try / except block
            n_thread_cmdline = plugin_args["n_procs"]
            if n_thread_cmdline > n_cpu:
                cprint(
                    msg=(
                        f"You are trying to run clinica with a number of threads ({n_thread_cmdline}) superior to your "
                        f"number of CPUs ({n_cpu})."
                    ),
                    lvl="warning",
                )
                ask_user = True
        except TypeError:
            cprint(
                msg=f"You did not specify the number of threads to run in parallel (--n_procs argument).",
                lvl="warning",
            )
            cprint(
                msg=(
                    f"Computation time can be shorten as you have {n_cpu} CPUs on this computer. "
                    f"We recommend using {n_cpu-1} threads."
                ),
                lvl="warning",
            )
            ask_user = True

        if ask_user:
            n_procs = click.prompt(
                text="How many threads do you want to use?",
                default=max(1, n_cpu - 1),
                show_default=True,
            )
            click.echo(
                f"Number of threads set to {n_procs}. You may set the --n_procs argument "
                "to disable this message for future calls."
            )
            if plugin_args:
                plugin_args["n_procs"] = n_procs
            else:
                plugin_args = {"n_procs": n_procs}

        return plugin_args

    def check_not_cross_sectional(self):
        """
        This function checks if the dataset is longitudinal. If it is cross
        sectional, clinica proposes to convert it in a clinica compliant form.

        author: Arnaud Marcoux
        """
        import sys
        from os import listdir
        from os.path import abspath, basename, dirname, isdir, join

        from clinica.utils.exceptions import ClinicaInconsistentDatasetError
        from clinica.utils.stream import cprint

        def _check_cross_subj(cross_subj: list) -> None:
            if len(cross_subj) > 0:
                raise ClinicaInconsistentDatasetError(cross_subj)

        def convert_cross_sectional(bids_in, bids_out, cross_subjects, long_subjects):
            """
            This function converts a cross-sectional-bids dataset into a
            longitudinal clinica-compliant dataset

            Args:
                bids_in: cross sectional bids dataset you want to convert
                bids_out: converted longitudinal bids dataset
                cross_subjects: list of subjects in cross sectional form
                (they need some adjustment)
                long_subjects: list of subjects in longitudinal form (they
                just need to be copied)

            Returns:
                nothing
            """
            from os import mkdir
            from os.path import exists, isfile
            from shutil import copy2, copytree

            def add_ses(f):
                """
                Use this function to transform a cross sectional filename into
                a longitudinal one.
                Examples:
                sub-ADNI001_scans.tsv -> sub-ADNI001_ses-M00_scans.tsv
                sub-023a_ses-M12_T1w.nii.gz -> sub-023a_ses-M12_T1w.nii.gz (no
                    modification done if filename already has a session)

                Args:
                    f: filename

                Returns:
                    filename with '_ses-M00_ added just after participant_id
                """
                import re

                # If filename contains ses-..., returns the original filename
                # Regex explication:
                # ^ start of string
                # ([a-zA-Z0-9]*) matches any number of characters from a to z,
                #       A to Z, 0 to 9, and store it in group(1)
                # (?!ses-[a-zA-Z0-9]) do not match if there is already a 'ses-'
                # (.*) catches the rest of the string
                m = re.search(r"(^sub-[a-zA-Z0-9]*)_(?!ses-[a-zA-Z0-9])(.*)", f)
                try:
                    return m.group(1) + "_ses-M00_" + m.group(2)
                except AttributeError:
                    # If something goes wrong, we return the original filename
                    return f

            def copy2_add_ses(src, dst):
                """
                copy2_add_ses calls copy2 function from shutil, but modifies
                the filename of the copied files if they match the regex
                template described in add_ses() function

                Args:
                    src: path to the file that needs to be copied
                    dst: original destination for the copied file

                Returns:
                    copy2 with modified filename
                """
                from os.path import basename, dirname, join
                from shutil import copy2

                dst_modified = join(dirname(dst), add_ses(basename(src)))
                return copy2(src, dst_modified)

            if not exists(bids_out):
                # Create the output folder if it does not exists yet
                mkdir(bids_out)

            # First part of the algorithm: deal with subjects that does not
            # have longitudinal (session) information
            for subj in cross_subjects:
                # Get list of of files/folders to copy. Remove hidden element
                # ( if they start with a dot '.')
                to_copy = [
                    f for f in listdir(join(bids_in, subj)) if not f.startswith(".")
                ]
                # Always check that folder are existing, otherwise an exception
                # is raised even though that does not prevent the function from
                # working
                if not exists(join(bids_out, subj)):
                    mkdir(join(bids_out, subj))
                if not exists(join(bids_out, subj, "ses-M00")):
                    mkdir(join(bids_out, subj, "ses-M00"))
                for el in to_copy:
                    path_el = join(bids_in, subj, el)
                    if not exists(join(bids_out, subj, "ses-M00", el)):
                        # If the element to copy is a folder...
                        if isdir(path_el):
                            # Copytree is used with a 'custom' copy function,
                            # to give a correct filemename to the files inside
                            # the copied folders
                            copytree(
                                path_el,
                                join(bids_out, subj, "ses-M00", basename(path_el)),
                                copy_function=copy2_add_ses,
                            )
                        # If the element to copy is a file...
                        elif isfile(path_el):
                            # Modify the filename with add_ses function
                            new_filename_wo_ses = add_ses(el)
                            copy2(
                                path_el,
                                join(bids_out, subj, "ses-M00", new_filename_wo_ses),
                            )
            # Second part of the algorithm: deal with subjects that do not
            # have the problem. We only xopy the content of the folder, and no
            # filename needs to be changed
            for su in long_subjects:
                # Do not copy hidden files
                to_copy = [
                    f for f in listdir(join(bids_in, su)) if not f.startswith(".")
                ]
                if not exists(join(bids_out, su)):
                    mkdir(join(bids_out, su))
                for el in to_copy:
                    path_el = join(bids_in, su, el)
                    if not exists(join(bids_out, su, el)):
                        # 2 possible cases: element to pcopy is a folder or a
                        # file
                        if isdir(path_el):
                            copytree(path_el, join(bids_out, su, basename(path_el)))
                        elif isfile(path_el):
                            copy2(path_el, join(bids_out, su))

        if self.bids_directory is not None:
            bids_dir = abspath(self.bids_directory)
            # Extract all subjects in BIDS directory: element must be a folder and
            # a name starting with 'sub-'
            all_subs = [
                f
                for f in listdir(bids_dir)
                if isdir(join(bids_dir, f)) and f.startswith("sub-")
            ]

            cross_subj, long_subj = detect_cross_sectional_and_longitudinal_subjects(
                all_subs, bids_dir
            )
            try:
                _check_cross_subj(cross_subj)
            except ClinicaInconsistentDatasetError as e:
                cprint(e, lvl="warning")
                proposed_bids = join(
                    dirname(bids_dir), basename(bids_dir) + "_clinica_compliant"
                )

                if not click.confirm(
                    "Do you want to proceed with the conversion in another "
                    "folder? (Your original BIDS folder will not be modified "
                    f"and the folder {proposed_bids} will be created.)"
                ):
                    click.echo("Clinica will now exit...")
                    sys.exit()
                else:
                    cprint("Converting cross-sectional dataset into longitudinal...")
                    convert_cross_sectional(
                        bids_dir, proposed_bids, cross_subj, long_subj
                    )
                    cprint(
                        f"Conversion succeeded. Your clinica-compliant dataset is located here: {proposed_bids}"
                    )

    @property
    def base_dir_was_specified(self):
        return self._base_dir_was_specified

    @property
    def is_built(self):
        return self._is_built

    @is_built.setter
    def is_built(self, value):
        self._is_built = value

    @property
    def overwrite_caps(self):
        return self._overwrite_caps

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters = value
        # Need to rebuild input, output and core nodes
        self.is_built = False
        self.init_nodes()

    @property
    def info(self):
        return self._info

    @info.setter
    def info(self, value):
        self._info = value

    @property
    def input_node(self):
        return self._input_node

    @property
    def output_node(self):
        return self._output_node

    @property
    def bids_directory(self):
        return self._bids_directory

    @property
    def caps_directory(self):
        return self._caps_directory

    @property
    def subjects(self):
        return self._subjects

    @subjects.setter
    def subjects(self, value):
        self._subjects = value
        self.is_built = False

    @property
    def sessions(self):
        return self._sessions

    @sessions.setter
    def sessions(self, value):
        self._sessions = value
        self.is_built = False

    @property
    def tsv_file(self):
        return self._tsv_file

    @property
    def info_file(self):
        return self._info_file

    @abc.abstractmethod
    def build_core_nodes(self):
        """Builds the Pipeline's core nodes.

        This method should use and connect to the `Pipeline.input_node` one
        or more core `Node`s. The outputs of the core processing should then be
        connected to the `Pipeline.output_node`.
        """

    @abc.abstractmethod
    def build_input_node(self):
        """Builds the Pipeline's input data stream node.

        Warnings:
            This method does not modify the `Pipeline.input_node` (see the
            notes about the global architecture in the class documentation).
        """

    @abc.abstractmethod
    def build_output_node(self):
        """Builds the Pipeline's output data stream node.

        Warnings:
            This method does not modify the `Pipeline.output_node` (see the
            notes about the global architecture in the class documentation).
        """

    @abc.abstractmethod
    def get_input_fields(self):
        """Lists the input fields of the Pipeline.

        Returns:
            A list of strings defining the fields of the `IdentityInterface`
            of the `Pipeline.input_node`.
        """

    @abc.abstractmethod
    def get_output_fields(self):
        """Lists the output fields of the Pipeline.

        Returns:
            A list of strings defining the fields of the `IdentityInterface`
            of the `Pipeline.output_node`.
        """

    @abc.abstractmethod
    def check_custom_dependencies(self):
        """Check dependencies provided by the developer."""

    @abc.abstractmethod
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""



from os import PathLike
from typing import BinaryIO, Dict, Iterable, List, Optional, Tuple, Union

from pandas import DataFrame


def find_clinical_data(clinical_data_directory: PathLike) -> Optional[DataFrame]:
    from pathlib import Path

    from pandas import read_excel

    dataframe = None
    for path in Path(clinical_data_directory).rglob("*.xlsx"):
        try:
            dataframe = (
                read_excel(path, index_col=[0, 4])  # noqa
                .rename(columns=lambda x: x.lower().replace(" ", "_"))
                .rename_axis(index=lambda x: x.lower().replace(" ", "_"))
                .convert_dtypes()
                .sort_index()
            )
        except (ValueError, IndexError):
            continue

    return dataframe


def read_clinical_data(clinical_data_directory: PathLike) -> DataFrame:
    import pandas as pd

    dataframe = find_clinical_data(clinical_data_directory)

    if dataframe is None:
        raise FileNotFoundError("Clinical data not found")

    # Compute participant and session IDs.
    dataframe = dataframe.rename_axis(
        index={"loni_id": "participant_id", "visit_number": "session_id"}
    )
    dataframe.index = dataframe.index.map(
        lambda x: (f"sub-NIFD{x[0].replace('_', '')}", f"ses-M{(6 * (x[1] - 1)):02d}")
    )

    # Keep relevant columns and rename them.
    dataframe = (
        dataframe[["dx", "site", "education", "race", "cdr_box_score", "mmse_tot"]]
        .rename(columns={"dx": "diagnosis", "cdr_box_score": "cdr", "mmse_tot": "mmse"})
        .astype(
            dtype={
                "diagnosis": pd.CategoricalDtype(
                    ["BV", "CON", "L_SD", "PATIENT (OTHER)", "PNFA", "SV"]
                ),
                "site": pd.CategoricalDtype(["UCSF", "MAYO", "MGH"]),
                "education": pd.Int64Dtype(),
                "race": pd.Int64Dtype(),
                "cdr": pd.Float64Dtype(),
                "mmse": pd.Float64Dtype(),
            }
        )
        .replace({"education": {99: pd.NA}, "race": {50: pd.NA, 99: pd.NA}})
    )

    # Keep positive MMSE values only.
    dataframe.mmse = dataframe.mmse.mask(dataframe.mmse < 0)

    return dataframe


def find_collection_data(imaging_data_directory: PathLike) -> Optional[DataFrame]:
    from pathlib import Path

    from pandas import read_csv

    dataframe = None
    for path in Path(imaging_data_directory).rglob("*.csv"):
        try:
            dataframe = (
                read_csv(
                    path,
                    index_col="Image Data ID",
                    parse_dates=["Acq Date"],
                )
                .rename(columns=lambda x: x.lower().replace(" ", "_"))
                .rename_axis(index=lambda x: x.lower().replace(" ", "_"))
                .convert_dtypes()
                .sort_index()
            )
        except (ValueError, IndexError):
            continue

    return dataframe


def find_imaging_data(imaging_data_directory: PathLike) -> Iterable[Tuple[str, str]]:
    import re
    from pathlib import Path

    # Pattern for extracting the image data ID from the NIFD files.
    pattern = re.compile(r"(I\d{6})")

    def find_files(in_: PathLike) -> Iterable[Path]:
        return filter(lambda x: x.is_file(), Path(in_).rglob("NIFD*.*"))

    def extract_id_with_dir(files: Iterable[Path]) -> Tuple[str, str]:
        for f in files:
            found = pattern.search(f.name)
            if found:
                yield found.group(1), str(f.parent)

    for image_data_id, source_dir in set(
        sorted(extract_id_with_dir(find_files(imaging_data_directory)))
    ):
        yield image_data_id, source_dir


def read_imaging_data(imaging_data_directory: PathLike) -> DataFrame:
    from pandas import DataFrame

    try:
        imaging_data = DataFrame.from_records(
            data=find_imaging_data(imaging_data_directory),
            columns=["image_data_id", "source_dir"],
            index="image_data_id",
        ).convert_dtypes()
    except TypeError:
        raise FileNotFoundError("No imaging data found")

    collection_data = find_collection_data(imaging_data_directory)

    if collection_data is None:
        raise FileNotFoundError("No collection data found")

    return collection_data.join(imaging_data)


def parse_mri_description(description: str) -> Optional[Dict[str, Optional[str]]]:
    description = description.lower().replace("-", "")

    if "mprage" in description:
        return {
            "datatype": "anat",
            "suffix": "T1w",
            "trc_label": None,
            "rec_label": None,
        }
    elif "flair" in description:
        return {
            "datatype": "anat",
            "suffix": "FLAIR",
            "trc_label": None,
            "rec_label": None,
        }
    elif "t2" in description:
        return {
            "datatype": "anat",
            "suffix": "T2w",
            "trc_label": None,
            "rec_label": None,
        }
    elif "asl" in description:
        return {
            "datatype": "anat",
            "suffix": "PDw",
            "trc_label": None,
            "rec_label": None,
        }
    elif any([x in description for x in ["mt1", "gradwarp", "n3m"]]):
        return {
            "datatype": "anat",
            "suffix": "T1w",
            "trc_label": None,
            "rec_label": None,
        }
    else:
        return None


def parse_pet_description(description: str) -> Optional[Dict[str, str]]:
    import re

    match = re.search(r"3D:(\w+):(\w+)", description)

    if match:
        return {
            "datatype": "pet",
            "suffix": "pet",
            "trc_label": "11CPIB" if "PIB" in match.group(1) else "18FFDG",
            "rec_label": "IR" if "IR" in match.group(2) else "RP",
        }
    else:
        return None


def parse_preprocessing(description: str) -> dict:
    description = description.lower()

    return {
        "gradwarp": any([x in description for x in ["gradwarp", "dis3d"]]),
        "n3": "n3m" in description,
    }


def dataset_to_bids(
    imaging_data: DataFrame,
    clinical_data: Optional[DataFrame] = None,
) -> Tuple[DataFrame, DataFrame, DataFrame]:
    from pandas import Series

    # Parse preprocessing information from scan descriptions.
    preprocessing = imaging_data.description.apply(parse_preprocessing).apply(Series)

    # Parse BIDS entities from scan descriptions.
    bids = (
        imaging_data.apply(
            lambda x: parse_pet_description(x.description)
            if x.modality == "PET"
            else parse_mri_description(x.description),
            axis=1,
        )
        .dropna()
        .apply(Series)
    )

    # Compute quality metric for each scan:
    # - MRI: Applied preprocessing (0: None, 1: GradWarp, 2: N3)
    # - PET: Reconstruction method (0: Fourier, 1: Iterative)
    quality = (
        preprocessing.sum(axis=1)
        + bids.rec_label.apply(lambda x: 1 if x == "IR" else 0)
    ).rename("quality")

    # Select one scan per BIDS modality based on quality metric.
    subset = ["subject", "visit", "datatype", "suffix", "trc_label"]
    scans = (
        bids.join(quality)
        .join(imaging_data)
        .sort_values(by=subset + ["quality"])
        .drop_duplicates(subset=subset, keep="last")
        .drop(columns="quality")
    )

    # Compute the BIDS-compliant participant, session and scan IDs.
    scans = scans.assign(
        participant_id=lambda df: df.subject.apply(
            lambda x: f"sub-NIFD{x.replace('_', '')}"
        ),
        session_id=lambda df: df.visit.apply(lambda x: f"ses-M{(6 * (x - 1)):02d}"),
        filename=lambda df: df.apply(
            lambda x: f"{x.participant_id}/{x.session_id}/{x.datatype}/"
            f"{x.participant_id}_{x.session_id}"
            f"{'_trc-' + x.trc_label if x.trc_label else ''}"
            f"{'_rec-' + x.rec_label if x.rec_label else ''}"
            f"_{x.suffix}.nii.gz",
            axis=1,
        ),
    )

    # Prepare subjects manifest.
    subjects = (
        scans[["participant_id", "session_id", "group", "sex", "age"]]
        .sort_values(by=["participant_id", "session_id"])
        .drop(columns="session_id")
        .drop_duplicates(subset="participant_id")
        .set_index(["participant_id"], verify_integrity=True)
        .sort_index()
    ).join(
        clinical_data.xs("ses-M00", level="session_id")[
            ["diagnosis", "site", "education", "race"]
        ]
    )

    # Prepare sessions manifest.
    sessions = (
        scans[["participant_id", "session_id", "acq_date", "age"]]
        .rename(columns={"acq_date": "date"})
        .drop_duplicates()
        .set_index(["participant_id", "session_id"], verify_integrity=True)
        .sort_index()
    ).join(clinical_data[["cdr", "mmse"]])

    # Prepare scans manifest.
    scans = scans[["filename", "source_dir", "format"]].set_index(
        "filename", verify_integrity=True
    )

    return subjects, sessions, scans


def write_to_tsv(dataframe: DataFrame, buffer: Union[PathLike, BinaryIO]) -> None:
    # Save dataframe as a BIDS-compliant TSV file.
    dataframe.to_csv(buffer, sep="\t", na_rep="n/a", date_format="%Y-%m-%d")


def convert_dicom(sourcedata_dir: PathLike, bids_filename: PathLike) -> None:
    from pathlib import PurePath

    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_utils import run_dcm2niix

    output_fmt = str(PurePath(bids_filename).name).replace(".nii.gz", "")
    output_dir = str(PurePath(bids_filename).parent)

    # Ensure output directory is empty.
    fs = LocalFileSystem()
    if fs.exists(output_dir):
        fs.rm(output_dir, recursive=True)
    fs.makedirs(output_dir)

    # Run conversion with dcm2niix with anonymization and maximum compression.
    run_dcm2niix(
        input_dir=str(sourcedata_dir),
        output_dir=output_dir,
        output_fmt=output_fmt,
        compress=True,
        bids_sidecar=True,
    )


def install_nifti(sourcedata_dir: PathLike, bids_filename: PathLike) -> None:
    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)
    source_file = fs.open(fs.ls(str(sourcedata_dir))[0], mode="rb")
    target_file = fs.open(str(bids_filename), mode="wb", compression="gzip")

    with source_file as sf, target_file as tf:
        tf.write(sf.read())


def write_bids(
    to: PathLike,
    participants: DataFrame,
    sessions: DataFrame,
    scans: DataFrame,
) -> List[PathLike]:
    from pathlib import PurePath

    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription

    to = PurePath(to)
    fs = LocalFileSystem(auto_mkdir=True)

    # Ensure BIDS hierarchy is written first.
    with fs.transaction:
        with fs.open(
            str(to / "dataset_description.json"), "w"
        ) as dataset_description_file:
            BIDSDatasetDescription(name="NIFD").write(to=dataset_description_file)
        with fs.open(str(to / "participants.tsv"), "w") as participant_file:
            write_to_tsv(participants, participant_file)

        for participant_id, sessions_group in sessions.groupby("participant_id"):
            participant_id = str(participant_id)
            sessions_group = sessions_group.droplevel("participant_id")
            sessions_filepath = to / participant_id / f"{participant_id}_sessions.tsv"
            with fs.open(str(sessions_filepath), "w") as sessions_file:
                write_to_tsv(sessions_group, sessions_file)

    # Perform import of imaging data next.
    for filename, metadata in scans.iterrows():
        filename = str(filename)
        if metadata.format == "DCM":
            convert_dicom(
                sourcedata_dir=metadata.source_dir, bids_filename=to / filename
            )
        else:
            install_nifti(
                sourcedata_dir=metadata.source_dir, bids_filename=to / filename
            )

    return scans.index.to_list()



"""This file contains functions for loading data from disk
and performing some checks on them.
"""
from os import PathLike
from pathlib import Path
from typing import Dict, Tuple

import numpy as np
import pandas as pd
from nilearn.surface import Mesh

TSV_FIRST_COLUMN = "participant_id"
TSV_SECOND_COLUMN = "session_id"


def _read_and_check_tsv_file(tsv_file: PathLike) -> pd.DataFrame:
    """This function reads the TSV file provided and performs some basic checks.

    Parameters
    ----------
    tsv_file : PathLike
        Path to the TSV file to open.

    Returns
    -------
    tsv_data : pd.DataFrame
        DataFrame obtained from the file.
    """
    try:
        return pd.read_csv(tsv_file, sep="\t").set_index(
            [TSV_FIRST_COLUMN, TSV_SECOND_COLUMN]
        )
    except FileNotFoundError:
        raise FileNotFoundError(f"File {tsv_file} does not exist.")
    except KeyError:
        raise ValueError(
            f"The TSV data should have at least two columns: {TSV_FIRST_COLUMN} and {TSV_SECOND_COLUMN}"
        )


def _get_t1_freesurfer_custom_file_template(base_dir: PathLike) -> str:
    """Returns a Template for the path to the desired surface file.

    Parameters
    ----------
    base_dir : PathLike
        Base directory to seach for the template.

    Returns
    -------
    template_path : str
        Path to the t1 freesurfer template.
    """
    return str(base_dir) + (
        "/%(subject)s/%(session)s/t1/freesurfer_cross_sectional/%(subject)s_%(session)s"
        "/surf/%(hemi)s.thickness.fwhm%(fwhm)s.fsaverage.mgh"
    )


def _build_thickness_array(
    input_dir: PathLike,
    surface_file: str,
    df: pd.DataFrame,
    fwhm: float,
) -> np.ndarray:
    """This function builds the cortical thickness array.

    Parameters
    ----------
    input_dir : PathLike
        Input directory.

    surface_file : str
        Template for the path to the surface file of interest.

    df : pd.DataFrame
        Subjects DataFrame.

    fwhm : float
        Smoothing parameter only used to retrieve the right surface file.

    Returns
    -------
    thickness : np.ndarray
        Cortical thickness. Hemispheres and subjects are stacked.
    """
    from nibabel.freesurfer.mghformat import load

    thickness = []
    for idx, row in df.iterrows():
        subject = row[TSV_FIRST_COLUMN]
        session = row[TSV_SECOND_COLUMN]
        parts = []
        for hemi in ["lh", "rh"]:
            query = {"subject": subject, "session": session, "fwhm": fwhm, "hemi": hemi}
            parts.append(
                load(str(Path(input_dir) / Path(surface_file % query))).get_fdata()
            )
        combined = np.vstack(parts)
        thickness.append(combined.flatten())
    thickness = np.vstack(thickness)
    if thickness.shape[0] != len(df):
        raise ValueError(
            f"Unexpected shape for thickness array : {thickness.shape}. "
            f"Expected {len(df)} rows."
        )
    return thickness


def _get_average_surface(fsaverage_path: PathLike) -> Tuple[Dict, Mesh]:
    """This function extracts the average surface and the average mesh
    from the path to the fsaverage templates.

    .. note::

        Note that the average surface is returned as a dictionary
        with 'coord' and 'tri' as keys, while the average mesh is
        returned as a Nilearn Mesh object (basically a NamedTuple
        with 'coordinates' and 'faces' attributes). The surface
        isn't returned as a Nilearn Surface object for compatibility
        with BrainStats.

    .. warning::

        There is an issue with faces having  a value of 0 as index.
        This is most likely a bug in BrainStat as MATLAB indexing
        starts at 1 while Python starts at zero.

    Parameters
    ----------
    fsaverage_path : PathLike
        Path to the fsaverage templates.

    Returns
    -------
    average_surface : dict
        Average surface as a dictionary for BrainStat compatibility.

    average_mesh : nilearn.surface.Mesh
        Average mesh as a Nilearn Mesh object.
    """
    import copy

    from nilearn.surface import Mesh, load_surf_mesh

    meshes = [
        load_surf_mesh(str(fsaverage_path / Path(f"{hemi}.pial")))
        for hemi in ["lh", "rh"]
    ]
    coordinates = np.vstack([mesh.coordinates for mesh in meshes])
    faces = np.vstack(
        [meshes[0].faces, meshes[1].faces + meshes[0].coordinates.shape[0]]
    )
    average_mesh = Mesh(
        coordinates=coordinates,
        faces=copy.deepcopy(faces),
    )
    ##################
    # UGLY HACK !!! Need investigation
    ##################
    # Uncomment the following line if getting an error
    # with negative values in bincount in Brainstat.
    # Not sure, but might be a bug in BrainStat...
    #
    faces += 1
    #################
    average_surface = {
        "coord": coordinates,
        "tri": faces,
    }
    return average_surface, average_mesh



import json
import os
from pathlib import Path


def build_bids_directory(directory: os.PathLike, subjects_sessions: dict) -> None:
    """Build a fake BIDS dataset at the specified location following the
    specified structure.

    Parameters
    ----------
    directory : PathLike
        Path to folder where the fake BIDS dataset should be created.

    subjects_sessions : Dict
        Dictionary containing the subjects and their associated sessions.

    Notes
    -----
    This function is a simple prototype for creating fake datasets for testing.
    It only adds (for now...) T1W nifti images for all subjects and sessions.
    """
    directory = Path(directory)
    data_types = {"anat"}
    suffixes = {"T1w", "flair"}
    extensions = {"nii.gz"}
    with open(directory / "dataset_description.json", "w") as fp:
        json.dump({"Name": "Example dataset", "BIDSVersion": "1.0.2"}, fp)
    for sub, sessions in subjects_sessions.items():
        (directory / sub).mkdir()
        for ses in sessions:
            (directory / sub / ses).mkdir()
            for data_type in data_types:
                (directory / sub / ses / data_type).mkdir()
                for suffix in suffixes:
                    for extension in extensions:
                        (
                            directory
                            / sub
                            / ses
                            / data_type
                            / f"{sub}_{ses}_{suffix}.{extension}"
                        ).touch()


def build_caps_directory(directory: os.PathLike, configuration: dict) -> None:
    """Build a fake CAPS dataset at the specified location following the
    specified structure.

    Parameters
    ----------
    directory : PathLike
        Path to folder where the fake CAPS dataset should be created.

    configuration : Dict
        Dictionary containing the configuration for building the fake CAPS
        directory. It should have the following structure:
            - "groups": ["group_labels"...]
            - "pipelines": ["pipeline_names"...]
            - "subjects": {"subject_labels": ["session_labels"...]}.

    Notes
    -----
    This function is a simple prototype for creating fake datasets for testing.
    """
    directory = Path(directory)
    _build_groups(directory, configuration)
    _build_subjects(directory, configuration)


def _build_groups(directory: Path, configuration: dict) -> None:
    """Build a fake groups file structure in a CAPS directory if requested."""
    if "groups" in configuration:
        (directory / "groups").mkdir()
        for group_label in configuration["groups"]:
            (directory / "groups" / f"group-{group_label}").mkdir()
            for pipeline in configuration["pipelines"]:
                (directory / "groups" / f"group-{group_label}" / pipeline).mkdir()
                (
                    directory
                    / "groups"
                    / f"group-{group_label}"
                    / pipeline
                    / f"group-{group_label}_template.nii.gz"
                ).touch()


def _build_subjects(directory: Path, configuration: dict) -> None:
    """Build a fake subjects file structure in a CAPS directory if requested."""
    if "subjects" in configuration:
        (directory / "subjects").mkdir()
        for sub, sessions in configuration["subjects"].items():
            (directory / "subjects" / sub).mkdir()
            for ses in sessions:
                (directory / "subjects" / sub / ses).mkdir()
                for pipeline in configuration["pipelines"]:
                    (directory / "subjects" / sub / ses / pipeline).mkdir()
                    if pipeline == "t1_linear":
                        _build_t1_linear(directory, sub, ses)
                    if pipeline == "t1":
                        _build_t1(directory, sub, ses, configuration)


def _build_t1_linear(directory: Path, sub: str, ses: str) -> None:
    """Build a fake t1-linear file structure in a CAPS directory."""
    (
        directory
        / "subjects"
        / sub
        / ses
        / "t1_linear"
        / f"{sub}_{ses}_T1w_space-MNI152NLin2009cSym_res-1x1x1_T1w.nii.gz"
    ).touch()


def _build_t1(directory: Path, sub: str, ses: str, configuration: dict) -> None:
    """Build a fake t1 file structure in a CAPS directory."""
    (directory / "subjects" / sub / ses / "t1" / "spm").mkdir()
    for folder in ["dartel", "segmentation"]:
        (directory / "subjects" / sub / ses / "t1" / "spm" / folder).mkdir()
    for group_label in configuration["groups"]:
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "dartel"
            / f"group-{group_label}"
        ).mkdir()
    (
        directory
        / "subjects"
        / sub
        / ses
        / "t1"
        / "spm"
        / "dartel"
        / f"group-{group_label}"
        / f"{sub}_{ses}_T1w_target-{group_label}_transformation-forward_deformation.nii.gz"
    ).touch()
    for tissue in ["csf", "graymatter", "whitematter"]:
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "dartel"
            / f"group-{group_label}"
            / f"{sub}_{ses}_T1w_segm-{tissue}_space-Ixi549Space_modulated-on_probability.nii.gz"
        ).touch()
    for folder in ["dartel_input", "native_space", "normalized_space"]:
        (
            directory / "subjects" / sub / ses / "t1" / "spm" / "segmentation" / folder
        ).mkdir()
    for tissue in ["csf", "graymatter", "whitematter"]:
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "segmentation"
            / "dartel_input"
            / f"{sub}_{ses}_T1w_segm-{tissue}_dartelinput.nii.gz"
        ).touch()
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "segmentation"
            / "native_space"
            / f"{sub}_{ses}_T1w_segm-{tissue}_probability.nii.gz"
        ).touch()
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "segmentation"
            / "normalized_space"
            / f"{sub}_{ses}_T1w_segm-{tissue}_space-Ixi549Space_modulated-off_probability.nii.gz"
        ).touch()


def rmtree(f: Path):
    if f.is_file():
        f.unlink()
    else:
        for child in f.iterdir():
            rmtree(child)
        f.rmdir()



from io import TextIOBase
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pandas as pd
from pandas import DataFrame, Series

import clinica.iotools.bids_utils as bids

PROTOCOL_TO_BIDS = {
    "3DFLAIR": {"datatype": "anat", "modality": "FLAIR"},
    "ADNI_1X-MPRAGE": {"datatype": "anat", "modality": "T1w"},
    "ADNI_2X-MPRAGE": {"datatype": "anat", "modality": "T1w"},
    "FDG": {
        "datatype": "pet",
        "modality": "pet",
        "trc_label": "18FFDG",
    },
    "FTP_75-105": {
        "datatype": "pet",
        "modality": "pet",
        "trc_label": "18FFTP",
    },
    "FTP_80-100": {
        "datatype": "pet",
        "modality": "pet",
        "trc_label": "18FFTP",
    },
    "PIB_40-60": {
        "datatype": "pet",
        "modality": "pet",
        "trc_label": "11CPIB",
    },
}


def source_participant_id_to_bids(dataframe: DataFrame) -> Series:
    # HABS participant format prefixed with `P_`
    return dataframe.source_participant_id.str.replace("P_", "sub-HABS", regex=False)


def source_session_id_to_bids(dataframe: DataFrame) -> Series:
    import re

    # HABS session format as `{years}.{months}`
    years_dot_months_pattern = r".*_(\d+).(\d+)"

    def replace_years_dot_months_session(match: re.Match) -> str:
        years = int(match.group(1)) - 1
        months = int(match.group(2))
        return f"ses-M{12 * years + months:02d}"

    # HABS undocumented format as `{months}m`
    months_only_pattern = r".*_(\d+)m.*"

    def replace_months_only_pattern(match: re.Match) -> str:
        months = int(match.group(1))
        return f"ses-M{months:02d}"

    return dataframe.source_session_id.str.replace(
        years_dot_months_pattern, replace_years_dot_months_session, regex=True
    ).str.replace(months_only_pattern, replace_months_only_pattern, regex=True)


def source_filename_to_bids(dataframe: DataFrame) -> Series:
    from pandas import notna

    cols = ["participant_id", "session_id", "datatype", "modality", "trc_label"]
    return dataframe[cols].apply(
        lambda row: (
            f"{row.datatype}/{row.participant_id}_{row.session_id}"
            f"{'_trc-'+row.trc_label if notna(row.trc_label) else ''}"
            f"_{row.modality}.nii.gz"
        ),
        axis=1,
    )


def find_clinical_data(sourcedata: Path):
    from csv import DictReader

    for f in sourcedata.rglob("*HABS_DataRelease*.csv"):
        header = DictReader(open(f, encoding="utf-8-sig")).fieldnames
        yield (
            f.name.split("_")[0],
            f.relative_to(sourcedata),
            {
                header[0]: "source_participant_id",
                header[1]: "source_session_id",
                header[2]: "date",
            },
        )


def read_clinical_data(path: Path, rename_columns: Dict[str, str]) -> DataFrame:
    from pandas import read_csv

    return (
        read_csv(path)
        .rename(columns=rename_columns)
        .assign(date=lambda df: pd.to_datetime(df.date))
        .assign(participant_id=source_participant_id_to_bids)
        .drop(columns="source_participant_id")
        .assign(session_id=source_session_id_to_bids)
        .drop(columns="source_session_id")
        .convert_dtypes()
        .set_index(["participant_id", "session_id", "date"], verify_integrity=True)
        .sort_index()
    )


def find_imaging_data(sourcedata: Path) -> List[Tuple[str, str]]:
    from zipfile import ZipFile

    ext = ".nii.gz"

    for z in sourcedata.rglob("*HABS_DataRelease*.zip"):
        yield [
            (str(z.relative_to(sourcedata)), f)
            for f in ZipFile(z).namelist()
            if f.endswith(ext)
        ]


def parse_imaging_data(paths: List[Tuple[str, str]]) -> Optional[DataFrame]:
    from pandas import concat, to_datetime

    dataframe = DataFrame.from_records(
        paths, columns=["source_zipfile", "source_filename"]
    )

    # Parse imaging metadata from file paths.
    pattern = (
        r"(?P<source_participant_id>P_\w{6})"
        r"_(?P<date>[\d-]+)"
        r"_(?P<source_session_id>HAB_\d{1}.\d{1,2})"
        r"_(?P<protocol>[\w-]+)"
    )

    dataframe = concat(
        [dataframe, dataframe["source_filename"].str.extract(pattern)],
        axis="columns",
    ).assign(date=lambda df: to_datetime(df.date))

    # Map protocol to BIDS entities.
    protocol_to_bids = DataFrame.from_dict(PROTOCOL_TO_BIDS, orient="index")

    dataframe = (
        dataframe.join(protocol_to_bids, on="protocol", how="left")
        .drop(columns="protocol")
        .dropna(subset=["datatype", "modality"])
    )

    if dataframe.empty:
        return

    # Compute BIDS participant ID, session ID and filename.
    dataframe = (
        dataframe.assign(participant_id=source_participant_id_to_bids)
        .drop(columns="source_participant_id")
        .assign(session_id=source_session_id_to_bids)
        .drop(columns="source_session_id")
        .set_index(
            ["participant_id", "session_id", "datatype", "modality", "trc_label"],
            verify_integrity=True,
        )
        .sort_index()
    )

    return dataframe


def write_to_tsv(dataframe: DataFrame, buffer: TextIOBase) -> None:
    """Save the input dataframe to a BIDS-compliant TSV format."""
    dataframe.to_csv(buffer, sep="\t", na_rep="n/a", date_format="%Y-%m-%d")


def install_nifti(zipfile: str, filename: str, bids_path: str) -> None:
    """Install a NIfTI file from a source archive to the target BIDS path."""
    import fsspec

    fo = fsspec.open(zipfile)
    fs = fsspec.filesystem("zip", fo=fo)
    with fsspec.open(bids_path, mode="wb") as f:
        f.write(fs.cat(filename))


def write_bids(
    sourcedata: Path,
    rawdata: Path,
    imaging_data: DataFrame,
    clinical_data: Dict[str, DataFrame],
):
    import fsspec
    from pandas import notna

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription

    participants = (
        clinical_data["Demographics"]
        .drop(columns="NP_Age")
        .rename(
            columns={
                "BiologicalSex": "sex",
                "YrsOfEd": "years_of_education",
            }
        )
        .xs("ses-M00", level="session_id")
        .reset_index(level="date", drop=True)
    ).rename(columns=str.lower)

    with fsspec.open(str(rawdata / "dataset_description.json"), mode="wt") as f:
        BIDSDatasetDescription(name="HABS").write(to=f)

    participants_file = rawdata / "participants.tsv"
    with fsspec.open(str(participants_file), mode="wb") as f:
        write_to_tsv(participants, f)

    sessions = (
        clinical_data["Demographics"][["NP_Age"]]
        .reset_index(level="date", drop=True)
        .rename(columns={"NP_Age": "age"})
        .join(
            clinical_data["ClinicalMeasures"]
            .rename(columns={"HABS_DX": "diagnostic"})
            .reset_index(level="date", drop=True)
        )
    ).rename(columns=str.lower)

    for participant_id, dataframe in sessions.groupby(["participant_id"]):
        sessions_file = rawdata / participant_id / f"{participant_id}_sessions.tsv"
        dataframe = dataframe.droplevel("participant_id")
        with fsspec.open(sessions_file, mode="wb") as f:
            write_to_tsv(dataframe, f)

    for grouped_by, dataframe in imaging_data.groupby(["participant_id", "session_id"]):
        participant_id, session_id = grouped_by

        bids_basedir = rawdata / participant_id / session_id
        bids_prefix = f"{participant_id}_{session_id}"

        dataframe = (
            dataframe.droplevel(level=["participant_id", "session_id"])
            .reset_index()
            .assign(
                filename=lambda df: df.apply(
                    lambda row: (
                        f"{row.datatype}/{bids_prefix}"
                        f"{'_trc-'+row.trc_label if notna(row.trc_label) else ''}"
                        f"_{row.modality}.nii.gz"
                    ),
                    axis="columns",
                )
            )
            .drop(columns=["datatype", "modality", "trc_label"])
            .set_index("filename", verify_integrity=True)
        )

        for filename, row in dataframe.iterrows():
            install_nifti(
                zipfile=str(sourcedata / row.source_zipfile),
                filename=row.source_filename,
                bids_path=str(bids_basedir / filename),
            )

        dataframe = dataframe.drop(columns=["source_zipfile", "source_filename"])

        with fsspec.open(bids_basedir / f"{bids_prefix}_scans.tsv", mode="wb") as f:
            write_to_tsv(dataframe, f)

    readme_data = {
        "link": "https://habs.mgh.harvard.edu",
        "desc": (
            "The overall goal of the Harvard Aging Brain Study (HABS) is to elucidate the earliest changes in "
            "molecular, functional and structural imaging markers that signal the transition from normal cognition to "
            "progressive cognitive decline along the trajectory of preclinical Alzheimer’s Disease."
        ),
    }
    bids.write_modality_agnostic_files(
        study_name="HABS", readme_data=readme_data, bids_dir=rawdata
    )



def get_luts():
    import os

    from clinica.utils.exceptions import ClinicaException

    try:
        # For aparc+aseg.mgz file:
        default = os.path.join(os.environ["FREESURFER_HOME"], "FreeSurferColorLUT.txt")
        # For aparc.a2009s+aseg.mgz file:
        a2009s = os.path.join(os.environ["FREESURFER_HOME"], "FreeSurferColorLUT.txt")

        # TODO: Add custom Lausanne2008 LUTs here.
    except KeyError:
        raise ClinicaException("Could not find FREESURFER_HOME environment variable.")
    return [default, a2009s]


def get_conversion_luts_offline():

    # TODO: use this function if no internet connect found in client (need to upload files to clinica repository)
    return


def get_conversion_luts():
    from os import pardir
    from os.path import abspath, dirname, join
    from pathlib import Path

    from clinica.utils.inputs import RemoteFileStructure, fetch_file
    from clinica.utils.stream import cprint

    root = dirname(abspath(join(abspath(__file__), pardir, pardir)))

    path_to_mappings = Path(root) / "resources" / "mappings"

    url_mrtrix = "https://raw.githubusercontent.com/MRtrix3/mrtrix3/master/share/mrtrix3/labelconvert/"

    fs_default = RemoteFileStructure(
        filename="fs_default.txt",
        url=url_mrtrix,
        checksum="6ee07088915fdbcf52b05147ddae86e5fcaf3efc63db5b0ba8f361637dfa11ef",
    )

    fs_a2009s = RemoteFileStructure(
        filename="fs_a2009s.txt",
        url=url_mrtrix,
        checksum="b472f09cfe92ac0b6694fb6b00a87baf15dd269566e4a92b8a151ff1080bf170",
    )

    ref_fs_default = path_to_mappings / Path(fs_default.filename)
    ref_fs_a2009 = path_to_mappings / Path(fs_a2009s.filename)

    if not (ref_fs_default.is_file()):
        try:
            ref_fs_default = fetch_file(fs_default, path_to_mappings)
        except IOError as err:
            cprint(
                msg=f"Unable to download required MRTRIX mapping (fs_default.txt) for processing: {err}",
                lvl="error",
            )
    if not (ref_fs_a2009.is_file()):
        try:
            ref_fs_a2009 = fetch_file(fs_a2009s, path_to_mappings)
        except IOError as err:
            cprint(
                msg=f"Unable to download required MRTRIX mapping (fs_a2009s.txt) for processing: {err}",
                lvl="error",
            )

    return [ref_fs_default, ref_fs_a2009]


def get_containers(subjects, sessions):
    import os

    return [
        os.path.join("subjects", subjects[i], sessions[i], "dwi")
        for i in range(len(subjects))
    ]


def get_caps_filenames(dwi_file: str):
    import re

    m = re.search(r"/(sub-[a-zA-Z0-9]+_ses-[a-zA-Z0-9]+.*)_preproc", dwi_file)
    if not m:
        raise ValueError(
            f"Input filename {dwi_file} is not in a CAPS compliant format."
        )
    source_file_caps = m.group(1)

    m = re.search(
        r"/(sub-[a-zA-Z0-9]+_ses-[a-zA-Z0-9]+.*)_space-[a-zA-Z0-9]+_preproc", dwi_file
    )
    if not m:
        raise ValueError(
            f"Input filename {dwi_file} is not in a CAPS compliant format."
        )
    source_file_bids = m.group(1)

    response = f"{source_file_caps}_model-CSD_responseFunction.txt"
    fod = f"{source_file_caps}_model-CSD_diffmodel.nii.gz"
    tracts = f"{source_file_caps}_model-CSD_tractography.tck"
    nodes = [
        f"{source_file_caps}_atlas-desikan_parcellation.nii.gz",
        f"{source_file_caps}_atlas-destrieux_parcellation.nii.gz",
    ]
    # TODO: Add custom Lausanne2008 node files here.
    connectomes = [
        f"{source_file_bids}_model-CSD_atlas-desikan_connectivity.tsv",
        f"{source_file_bids}_model-CSD_atlas-destrieux_connectivity.tsv",
    ]
    # TODO: Add custom Lausanne2008 connectome files here.

    return response, fod, tracts, nodes, connectomes


def print_begin_pipeline(in_bids_or_caps_file: str) -> None:
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    print_begin_image(get_subject_id(in_bids_or_caps_file))


def print_end_pipeline(in_bids_or_caps_file: str, final_file: str) -> None:
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(in_bids_or_caps_file))



from typing import Optional

import click

from clinica.pipelines import cli_param

pipeline_name = "compute-atlas"


@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.option.atlas_path
@cli_param.option.n_procs
def cli(
    caps_directory: str,
    atlas_path: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Projection of the results of t1-freesurfer on another atlas.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_FreeSurfer/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .t1_freeesurfer_atlas_pipeline import T1FreeSurferAtlas

    pipeline = T1FreeSurferAtlas(
        caps_directory=caps_directory,
        atlas_path=atlas_path,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



from os import PathLike

import click

from clinica.iotools.converters import cli_param


@click.command(name="ukb-to-bids")
@cli_param.dataset_directory
@cli_param.clinical_data_directory
@cli_param.bids_directory
def cli(
    dataset_directory: PathLike,
    clinical_data_directory: PathLike,
    bids_directory: PathLike,
) -> None:
    """UK Biobank to BIDS converter.

    Convert the imaging and clinical data of UK Biobank, located in DATASET_DIRECTORY and
    CLINICAL_DATA_DIRECTORY respectively, to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from clinica.iotools.bids_utils import _write_bidsignore
    from clinica.iotools.converters.ukb_to_bids.ukb_to_bids import convert_images
    from clinica.utils.check_dependency import check_dcm2niix
    from clinica.utils.stream import cprint

    check_dcm2niix()

    convert_images(dataset_directory, bids_directory, clinical_data_directory)
    _write_bidsignore(str(bids_directory))

    cprint("Conversion to BIDS succeeded.")


if __name__ == "__main__":
    cli()



from os import PathLike

from pydra.mark import annotate, task


@task
@annotate({"return": {"out_status": None}})
def task_volume_location_in_world_coordinate_system(
    nifti_input: PathLike,
    bids_dir: PathLike,
    modality: str = "t1w",
    skip_question: bool = False,
) -> bool:
    """Check if images are centered around the origin of the world coordinate

    Parameters
    ----
    nifti_input: PathLike
        Path to nifti file

    bids_dir: PathLike
        path to bids directory associated with this check

    modality: str, optional
        the modality of the image. Default="t1w".

    skip_question: bool, optional
        if True, assume answer is yes. Default=False.

    Returns
    -------
    bool
        True if they are centered, False otherwise

    Warns
    ------
    If volume is not centered on origin of the world coordinate system
    """
    from clinica.iotools.utils.data_handling import (
        check_volume_location_in_world_coordinate_system,
    )

    nifti_list = [str(nifti_input)]

    return check_volume_location_in_world_coordinate_system(
        nifti_list, bids_dir, modality, skip_question
    )


@task
@annotate({"return": {"out_list": list}})
def wrap_list(in_list: list):
    """Wraps a list inside another list to comply with Pydra Nipype1Task requirements

    Parameters
    ----------
    in_list: list
        The input list to wrap

    Returns
    -------
    list
        The wrapped list
    """
    return [in_list]



# coding: utf-8

"""Module for converting PIB PET of ADNI."""


def convert_adni_pib_pet(
    source_dir, csv_dir, dest_dir, conversion_dir, subjs_list=None, mod_to_update=False
):
    """Convert PIB PET images of ADNI into BIDS format.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        conversion_dir: path to the TSV files including the paths to original images
        subjs_list: subjects list
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from clinica.utils.stream import cprint

    if not subjs_list:
        adni_merge_path = path.join(csv_dir, "ADNIMERGE.csv")
        adni_merge = pd.read_csv(adni_merge_path, sep=",", low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint(
        f"Calculating paths of PIB PET images. Output will be stored in {conversion_dir}."
    )
    images = compute_pib_pet_paths(source_dir, csv_dir, subjs_list, conversion_dir)
    cprint("Paths of PIB PET images found. Exporting images into BIDS ...")
    paths_to_bids(images, dest_dir, "pib", mod_to_update=mod_to_update)
    cprint(msg="PIB PET conversion done.", lvl="debug")


def compute_pib_pet_paths(source_dir, csv_dir, subjs_list, conversion_dir):
    """Compute the paths to the PIB PET images and store them in a TSV file.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        find_image_path,
        get_images_pet,
    )
    from clinica.utils.pet import Tracer

    pet_pib_col = [
        "Phase",
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Original",
    ]
    pet_pib_df = pd.DataFrame(columns=pet_pib_col)
    pet_pib_dfs_list = []

    # Loading needed .csv files
    pibqc = pd.read_csv(path.join(csv_dir, "PIBQC.csv"), sep=",", low_memory=False)
    pet_meta_list = pd.read_csv(
        path.join(csv_dir, "PET_META_LIST.csv"), sep=",", low_memory=False
    )

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list["Subject"] == subj]

        if subject_pet_meta.empty:
            continue

        # QC for PIB PET images
        pet_qc_subj = pibqc[(pibqc.PASS == 1) & (pibqc.RID == int(subj[-4:]))]

        sequences_preprocessing_step = ["PIB Co-registered, Averaged"]
        subj_dfs_list = get_images_pet(
            subj,
            pet_qc_subj,
            subject_pet_meta,
            pet_pib_col,
            "PIB-PET",
            sequences_preprocessing_step,
            viscode_field="VISCODE",
        )
        if subj_dfs_list:
            pet_pib_dfs_list += subj_dfs_list

    if pet_pib_dfs_list:
        pet_pib_df = pd.concat(pet_pib_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = []

    # Removing known exceptions from images to convert
    if not pet_pib_df.empty:
        error_ind = pet_pib_df.index[
            pet_pib_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        pet_pib_df.drop(error_ind, inplace=True)

    images = find_image_path(pet_pib_df, source_dir, "PIB", "I", "Image_ID")
    images.to_csv(
        path.join(conversion_dir, f"{Tracer.PIB}_pet_paths.tsv"), sep="\t", index=False
    )

    return images



from cmath import nan
from os import PathLike
from typing import Dict, Iterable, List, Optional

from pandas import DataFrame, Series


def find_clinical_data(
    clinical_data_directory: PathLike,
) -> DataFrame:
    from pathlib import Path

    import pandas as pd

    from clinica.utils.stream import cprint

    # Read the xls
    try:
        image_data_file = list(Path(clinical_data_directory).glob("*.csv"))
    except StopIteration:
        raise FileNotFoundError("Clinical data file not found.")

    # Assign the dfs
    if len(image_data_file) == 0:
        raise FileNotFoundError("Clinical data not found or incomplete. Aborting")
    if len(image_data_file) > 1:
        raise ValueError("Too many data files found, expected one. Aborting.")
    df_clinical = pd.read_csv(
        str(image_data_file[0]),
        usecols=["eid", "31-0.0", "34-0.0", "21022-0.0", "21003-2.0", "21003-3.0"],
    )
    cprint(msg="All clinical data have been found", lvl="info")
    required_columns = [
        "eid",
        "31-0.0",
        "34-0.0",
        "21022-0.0",
        "21003-2.0",
        "21003-3.0",
    ]
    missing_columns = set(required_columns).difference(set(df_clinical.columns))
    if missing_columns:
        raise ValueError(f"Required columns {missing_columns} not found")
    df_clinical = df_clinical.rename(
        columns={
            "31-0.0": "sex_f31_0_0",
            "34-0.0": "year_of_birth_f34_0_0",
            "21022-0.0": "age_at_recruitment_f21022_0_0",
            "21003-2.0": "age_when_attended_assessment_centre_f21003_2_0",
            "21003-3.0": "age_when_attended_assessment_centre_f21003_3_0",
        }
    )
    return df_clinical


def read_imaging_data(imaging_data_directory: PathLike) -> DataFrame:
    from pathlib import Path

    import pandas as pd

    source_path_series_nifti = pd.Series(
        find_imaging_data(imaging_data_directory), name="source_path"
    )
    source_path_series_dicom = pd.Series(
        find_dicom_data(imaging_data_directory), name="source_path"
    )
    # Rationales for this choice of files can be found in the documentation:
    # https://github.com/aramis-lab/clinica/blob/dev/docs/Converters/UKBtoBIDS.md
    # list of the files we want to build the bids for each modality
    file_mod_list = [
        "T1.nii.gz",
        "T2_FLAIR.nii.gz",
        "AP.nii.gz",
        "PA.nii.gz",
        "rfMRI.dcm",
        "tfMRI.dcm",
        "SWI.nii.gz",
    ]

    dataframe_nifti = pd.DataFrame.from_records(
        source_path_series_nifti, columns=["source_zipfile", "source_filename"]
    )
    dataframe_nifti = dataframe_nifti[
        ~dataframe_nifti["source_filename"].str.contains("unusable")
    ]
    dataframe_dicom = pd.DataFrame.from_records(
        source_path_series_dicom, columns=["source_zipfile", "source_filename"]
    )
    dataframe = pd.concat([dataframe_nifti, dataframe_dicom], 0)
    filename = (
        dataframe["source_filename"]
        .apply(lambda x: Path(str(x)).name)
        .rename("filename")
    )
    dataframe = dataframe[filename.isin(file_mod_list)]
    filename = filename[filename.isin(file_mod_list)]
    if dataframe.empty:
        raise ValueError(
            f"No imaging data were found in the provided folder: {imaging_data_directory}, "
            "or they are not handled by Clinica. Please check your data."
        )
    split_zipfile = dataframe["source_zipfile"].str.split("_", expand=True)
    split_zipfile = split_zipfile.rename(
        {0: "source_id", 1: "modality_num", 2: "source_sessions_number"}, axis="columns"
    )  # .drop_duplicates()
    split_zipfile = split_zipfile.astype({"source_id": "int64"})

    df_source = pd.concat(
        [
            dataframe,
            filename,
            split_zipfile,
        ],
        axis=1,
    )
    return df_source


def find_dicom_data(path_to_source_data: PathLike) -> Iterable[PathLike]:
    """
    This function finds the paths to the dicoms, only for fMRI, since we use the niftis for the other modalities.
    More information here: https://github.com/aramis-lab/clinica/blob/dev/docs/Converters/UKBtoBIDS.md"""
    from pathlib import Path

    for z in Path(path_to_source_data).rglob("*.zip"):
        if str(z.name).split("_")[1] == "20217":
            yield [str(z.relative_to(path_to_source_data)), "fMRI/tfMRI.dcm"]
        elif str(z.name).split("_")[1] == "20225":
            yield [str(z.relative_to(path_to_source_data)), "fMRI/rfMRI.dcm"]


def find_imaging_data(path_to_source_data: PathLike) -> Iterable[PathLike]:
    from pathlib import Path
    from zipfile import ZipFile

    for z in Path(path_to_source_data).rglob("*.zip"):
        for f in ZipFile(z).namelist():
            if f.endswith(".nii.gz"):
                yield [str(z.relative_to(path_to_source_data)), f]


def intersect_data(df_source: DataFrame, df_clinical_data: DataFrame) -> DataFrame:
    """
    This function merges the two dataframes given as inputs based on the subject id
    """

    df_clinical = df_clinical_data.merge(
        df_source, how="inner", right_on="source_id", left_on="eid"
    )
    return df_clinical


def complete_clinical(df_clinical: DataFrame) -> DataFrame:
    """This function uses the existing data to create the columns needed for
    the bids hierarchy (subject_id, ses, age_at _sessions, ect.)"""
    import pandas as pd

    df_clinical = df_clinical.assign(
        participant_id=lambda df: ("sub-UKB" + df.source_id.astype("str"))
    )
    df_clinical = df_clinical.assign(
        sessions=lambda df: "ses-" + df.source_sessions_number.astype("str")
    )
    df_clinical = df_clinical.join(
        df_clinical.filename.map(
            {
                "T2_FLAIR.nii.gz": {
                    "datatype": "anat",
                    "modality": "FLAIR",
                    "suffix": "FLAIR",
                    "sidecars": ["T2_FLAIR/T2_FLAIR.json"],
                    "task": "",
                    "dir": "",
                },
                "T1.nii.gz": {
                    "datatype": "anat",
                    "modality": "T1W",
                    "suffix": "T1w",
                    "sidecars": ["T1/T1.json"],
                    "task": "",
                    "dir": "",
                },
                "AP.nii.gz": {
                    "datatype": "dwi",
                    "modality": "dwi",
                    "suffix": "dwi",
                    "sidecars": [
                        "dMRI/raw/AP.json",
                        "dMRI/raw/AP.bval",
                        "dMRI/raw/AP.bvec",
                    ],
                    "task": "",
                    "dir": "_dir-AP",
                },
                "PA.nii.gz": {
                    "datatype": "dwi",
                    "modality": "dwi",
                    "suffix": "dwi",
                    "sidecars": [
                        "dMRI/raw/PA.json",
                        "dMRI/raw/PA.bval",
                        "dMRI/raw/PA.bvec",
                    ],
                    "task": "",
                    "dir": "_dir-PA",
                },
                "rfMRI.dcm": {
                    "datatype": "func",
                    "modality": "rsfmri",
                    "suffix": "bold",
                    "sidecars": [],
                    "task": "_task-rest",
                    "dir": "",
                },
                "tfMRI.dcm": {
                    "datatype": "func",
                    "modality": "tfmri",
                    "suffix": "bold",
                    "sidecars": [],
                    "task": "_task-facesshapesemotion",
                    "dir": "",
                },
                "SWI.nii.gz": {
                    "datatype": "swi",
                    "modality": "swi",
                    "suffix": "swi",
                    "sidecars": [],
                    "task": "",
                    "dir": "",
                },
            }
        ).apply(pd.Series)
    )
    df_clinical = df_clinical.assign(year_of_birth=lambda df: df.year_of_birth_f34_0_0)

    df_clinical = df_clinical.assign(age=lambda df: df.age_at_recruitment_f21022_0_0)
    df_clinical = df_clinical.assign(
        age_at_first_session=lambda df: df.age_when_attended_assessment_centre_f21003_2_0
    )
    df_clinical["age_at_session"] = df_clinical.apply(
        lambda df: select_sessions(df), axis=1
    )
    df_clinical = df_clinical[df_clinical["age_at_session"].notna()]
    df_clinical = df_clinical.assign(
        sessions_month=lambda df: (df.age_at_session - df.age_at_first_session) * 12
    )
    df_clinical = df_clinical.assign(
        sessions=lambda df: df.sessions_month.map(lambda x: f"ses-M{int(x):03d}")
    )
    df_clinical = df_clinical.assign(
        bids_filename=lambda df: (
            "sub-UKB"
            + df.source_id.astype("str")
            + "_"
            + df.sessions.astype("str")
            + df.dir.astype("str")
            + df.task.astype("str")
            + "_"
            + df.suffix
        )
    )
    df_clinical = df_clinical.join(
        df_clinical.sex_f31_0_0.astype("str")
        .map(
            {
                "0": {"sex": "F"},
                "1": {"sex": "M"},
            }
        )
        .apply(pd.Series)
    )
    df_clinical = df_clinical.assign(
        bids_full_path=lambda df: df.participant_id
        + "/"
        + df.sessions
        + "/"
        + df.datatype
        + "/"
        + df.bids_filename
    )
    return df_clinical


def dataset_to_bids(df_clinical: DataFrame) -> Dict[str, DataFrame]:

    import os

    import pandas as pd

    # open the reference for building the tsvs:
    path_to_ref_csv = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
        "data",
        "ukb_ref.csv",
    )
    df_ref = pd.read_csv(path_to_ref_csv, sep=";")

    df_clinical = df_clinical.set_index(
        ["participant_id", "sessions", "modality", "bids_filename"],
        verify_integrity=True,
    )

    return {
        col: df_clinical.filter(items=list(df_ref[col]))
        for col in ["participants", "sessions", "scans"]
    }


def write_bids(
    to: PathLike,
    participants: DataFrame,
    sessions: DataFrame,
    scans: DataFrame,
    dataset_directory: PathLike,
) -> None:
    from pathlib import Path

    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
    from clinica.iotools.bids_utils import write_to_tsv

    to = Path(to)
    fs = LocalFileSystem(auto_mkdir=True)

    participants = participants.droplevel(
        ["sessions", "modality", "bids_filename"]
    ).drop_duplicates()

    # Ensure BIDS hierarchy is written first.
    with fs.transaction:
        with fs.open(
            str(to / "dataset_description.json"), "w"
        ) as dataset_description_file:
            BIDSDatasetDescription(name="UKB").write(to=dataset_description_file)
        with fs.open(str(to / "participants.tsv"), "w") as participant_file:
            write_to_tsv(participants, participant_file)

    for participant_id, data_frame in sessions.groupby(["participant_id"]):
        sessions = data_frame.droplevel(
            ["participant_id", "modality", "bids_filename"]
        ).drop_duplicates()

        sessions_filepath = to / str(participant_id) / f"{participant_id}_sessions.tsv"
        with fs.open(str(sessions_filepath), "w") as sessions_file:
            write_to_tsv(sessions, sessions_file)
    scans = scans.set_index(["bids_full_path"], verify_integrity=True)
    for bids_full_path, metadata in scans.iterrows():
        if metadata["modality_num"] != "20217" and metadata["modality_num"] != "20225":
            copy_file_to_bids(
                zipfile=str(dataset_directory) + "/" + metadata["source_zipfile"],
                filenames=[metadata["source_filename"]] + metadata["sidecars"],
                bids_path=to / bids_full_path,
            )
        else:
            convert_dicom_to_nifti(
                zipfiles=str(dataset_directory) + "/" + metadata["source_zipfile"],
                bids_path=to / bids_full_path,
            )
            if metadata["modality_num"] == "20217":
                import_event_tsv(bids_path=str(to))
    return


def copy_file_to_bids(zipfile: str, filenames: List[str], bids_path: str) -> None:
    """Install the requested files in the BIDS  dataset."""
    import fsspec

    fo = fsspec.open(zipfile)
    fs = fsspec.filesystem("zip", fo=fo)
    for filename in filenames:
        if fs.exists(filename):
            bids_path_extension = str(bids_path) + "." + (filename.split(".", 1)[1])
            with fsspec.open(bids_path_extension, mode="wb") as f:
                f.write(fs.cat(filename))


def convert_dicom_to_nifti(zipfiles: str, bids_path: str) -> None:
    """Install the requested files in the BIDS  dataset.
    First, the dicom is extracted in a temporary directory
    Second, the dicom extracted is converted in the right place using dcm2niix"""
    import json
    import os
    import subprocess
    import tempfile
    import zipfile
    from pathlib import PurePath

    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)

    zf = zipfile.ZipFile(zipfiles)
    try:
        os.makedirs(PurePath(bids_path).parent)
    except OSError:
        # Folder already created with previous instance
        pass
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        command = [
            "dcm2niix",
            "-w",
            "0",
        ]
        command += ["-9", "-z", "y"]
        command += ["-b", "y", "-ba", "y"]
        command += [tempdir]
        subprocess.run(command)
        fmri_image_path = PurePath(find_largest_imaging_data(tempdir))
        fs.copy(str(fmri_image_path), str(bids_path) + ".nii.gz")
        fs.copy(
            str(fmri_image_path.parent)
            + "/"
            + PurePath(PurePath(fmri_image_path.name).stem).stem
            + ".json",
            str(bids_path) + ".json",
        )
    # Add the taskname to the json
    with open(str(bids_path) + ".json", "r+") as f:
        json_file = json.load(f)
        json_file["TaskName"] = "facesshapesemotion"
        f.seek(0)
        json.dump(json_file, f, indent=4)
    return


def select_sessions(x: DataFrame) -> Optional[Series]:
    from clinica.utils.stream import cprint

    if (
        x["source_sessions_number"] == "2"
        and x.age_when_attended_assessment_centre_f21003_2_0 != nan
    ):
        return x.age_when_attended_assessment_centre_f21003_2_0
    elif (
        x["source_sessions_number"] == "2"
        and x.age_when_attended_assessment_centre_f21003_2_0 == nan
    ):
        cprint(
            msg=f"The subject {x.eid} doesn't have the age for the imaging session number one (age_when_attended_assessment_centre_f21003_2_0)."
            f"It will not be converted. To have it converted, please update your clinical data.",
            lvl="warning",
        )
        return None
    elif (
        x["source_sessions_number"] == "3"
        and x.age_when_attended_assessment_centre_f21003_3_0 != nan
    ):
        return x.age_when_attended_assessment_centre_f21003_3_0
    elif (
        x["source_sessions_number"] == "3"
        and x.age_when_attended_assessment_centre_f21003_3_0 == nan
    ):
        cprint(
            msg=f"The subject {x.eid} doesn't have the age for the imaging session number two (age_when_attended_assessment_centre_f21003_3_0)."
            f"It will not be converted. To have it converted, please update your clinical data.",
            lvl="warning",
        )
        return None


def import_event_tsv(bids_path: str) -> None:
    """Import the csv containing the events' information."""
    import os

    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)
    path_to_event_tsv = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))),
        "resources",
        "fmri",
        "task-facesshapesemotion_events.tsv",
    )

    bids_path_extension = str(bids_path) + "/" + "task-facesshapesemotion_events.tsv"
    fs.copy(path_to_event_tsv, bids_path_extension)
    return


def find_largest_imaging_data(path_to_source_data: PathLike) -> PathLike:
    import os
    from pathlib import Path

    weight = 0
    file_to_return = ""
    for z in Path(path_to_source_data).rglob("*.nii.gz"):
        if os.path.getsize(z) > weight:
            weight = os.path.getsize(z)
            file_to_return = z
    return Path(file_to_return)



"""This module contains utilities for longitudinal pipelines. See CAPS specifications for details about long ID."""


def get_long_id(list_session_id):
    """Extract longitudinal ID from a set of session IDs.

    This will create a unique identifier for a participant and its corresponding sessions. Sessions labels are sorted
    alphabetically before being merged in order to generate the longitudinal ID.

    Args:
        list_session_id (list[str]): List of session IDs
            (e.g. ["ses-M00"] or ["ses-M00", "ses-M18", "ses-M36"])

    Returns:
        Longitudinal ID (str)

    Example:
        >>> from clinica.utils.longitudinal import get_long_id
        >>> get_long_id(['ses-M00'])
        'long-M00'
        >>> get_long_id(['ses-M00', 'ses-M18', 'ses-M36'])
        'long-M00M18M36'
        >>> get_long_id(['ses-M18', 'ses-M36', 'ses-M00']) # Session IDs do not need to be sorted
        'long-M00M18M36'
    """
    sorted_list = sorted(list_session_id)
    list_session_label = [session_id[4:] for session_id in sorted_list]
    long_id = "long-" + "".join(list_session_label)

    return long_id


def get_participants_long_id(list_participant_id, list_session_id):
    """Extract list of longitudinal IDs from a set of participant and session IDs.

    Example:
        >>> from clinica.utils.longitudinal import get_participants_long_id
        >>> get_participants_long_id(['sub-CLNC01', 'sub-CLNC01', 'sub-CLNC02'], ['ses-M00', 'ses-M18', 'ses-M00'])
        ['long-M00M18', 'long-M00M18', 'long-M00']
    """
    from .participant import get_unique_subjects

    unique_subject_list, per_subject_session_list = get_unique_subjects(
        list_participant_id, list_session_id
    )

    list_long_id = []
    for i in range(0, len(unique_subject_list)):
        list_long_id = list_long_id + [get_long_id(per_subject_session_list[i])] * len(
            per_subject_session_list[i]
        )

    return list_long_id


def save_long_id(list_session_id, output_dir, file_name=None):
    """Save long ID to `file_name`."""
    import os

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    long_id = get_long_id(list_session_id)
    file_name = file_name or f"{long_id}_sessions.tsv"
    sessions_tsv = open(os.path.join(output_dir, file_name), "w")
    sessions_tsv.write("session_id\n")

    for session_id in sorted(list_session_id):
        sessions_tsv.write(session_id + "\n")

    sessions_tsv.close()


def read_sessions(caps_dir, participant_id, long_id):
    """Extract sessions IDs from `caps_dir`/subjects/`participant_id`/`long_id`/`long_id`_sessions.tsv."""
    import os

    import pandas

    from clinica.utils.exceptions import ClinicaException

    sessions_file = os.path.join(
        os.path.expanduser(caps_dir),
        "subjects",
        participant_id,
        long_id,
        f"{long_id}_sessions.tsv",
    )
    if not os.path.isfile(sessions_file):
        raise ClinicaException(
            "The TSV file with sessions associated "
            f"to {participant_id} for longitudinal ID {long_id} is missing "
            f"(expected path: {sessions_file})."
        )
    ss_df = pandas.read_csv(sessions_file, sep="\t")
    if "session_id" not in list(ss_df.columns.values):
        raise ClinicaException(
            "The TSV file does not contain session_id column "
            f"(path: {sessions_file})."
        )

    return list(ss_df.session_id)



"""This module contains utilities to handle atlases in Clinica.

An atlas is currently defined by its name and a set of labels in a template space.

This current implementation has some drawbacks:
- Atlas is misleading: it is only a set of labels in a template space
- This implementation can not handle case where parcellation is in several template space
(e.g. `MNI152NLin2009cAsym` and `MNI152NLin6Asym`, see TemplateFlow for example)

Either a refactoring of this module or the use of an external API
(e.g. TemplateFlow - https://www.templateflow.org/) needs to be considered.
"""

import abc

T1_VOLUME_ATLASES = [
    "AAL2",
    "AICHA",
    "Hammers",
    "LPBA40",
    "Neuromorphometrics",
]

PET_VOLUME_ATLASES = [
    "AAL2",
    "AICHA",
    "Hammers",
    "LPBA40",
    "Neuromorphometrics",
]

DWI_DTI_ATLASES = [
    "JHUDTI81",
    "JHUTract0",
    "JHUTract25",
]

VOLUME_ATLASES = list(set(T1_VOLUME_ATLASES + PET_VOLUME_ATLASES + DWI_DTI_ATLASES))


class AtlasAbstract:
    """Abstract class for Atlas handling.

    Naming convention for children classes of AtlasAbstract:
    <name_atlas>[<resolution>][<map>]
    """

    __metaclass__ = abc.ABCMeta

    @staticmethod
    @abc.abstractmethod
    def get_name_atlas():
        """Return the name of the atlas (as defined in BIDS/CAPS specifications)."""

    def get_spatial_resolution(self):
        """Return the spatial resolution of the atlas (in format "XxXxX" e.g. 1.5x1.5x1.5)."""
        import nibabel as nib

        img_labels = nib.load(self.get_atlas_labels())
        voxels_labels = img_labels.header.get_zooms()
        # Will display integers without decimals
        if int(voxels_labels[0]) == voxels_labels[0]:
            s_x = str(int(voxels_labels[0]))
        else:
            s_x = str(voxels_labels[0])
        if int(voxels_labels[1]) == voxels_labels[1]:
            s_y = str(int(voxels_labels[1]))
        else:
            s_y = str(voxels_labels[1])
        if int(voxels_labels[2]) == voxels_labels[2]:
            s_z = str(int(voxels_labels[2]))
        else:
            s_z = str(voxels_labels[2])

        return f"{s_x}x{s_y}x{s_z}"

    @staticmethod
    @abc.abstractmethod
    def get_atlas_labels():
        """Return the image with the different labels/ROIs."""

    @staticmethod
    @abc.abstractmethod
    def get_tsv_roi():
        """Return the TSV file containing the ROI (regions of interest) of the atlas."""

    def get_index(self):
        import nibabel as nib
        import numpy as np

        img_labels = nib.load(self.get_atlas_labels())
        img_labels = img_labels.get_fdata(dtype="float32")
        labels = list(set(img_labels.ravel()))
        index_vector = np.zeros(len(labels))
        for index, n in enumerate(labels):
            index_vector[index] = index
        return index_vector


class JHUDTI811mm(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "JHUDTI81"

    @staticmethod
    def get_atlas_labels():
        import os

        import nipype.interfaces.fsl as fsl

        from .check_dependency import check_environment_variable
        from .inputs import _sha256

        fsl_dir = check_environment_variable("FSLDIR", "FSL")
        atlas_labels = os.path.join(
            fsl_dir, "data", "atlases", "JHU", "JHU-ICBM-labels-1mm.nii.gz"
        )

        # Adding checksum for updated file with version 6.0.5 of fsl
        fsl_atlas_checksums = {
            "old": "fac584ec75ff2a8631710d3345df96733ed87d9bde3387f5b462f8d22914ed69",
            "new": "3c3f5d2f1250a3df60982acff35a75b99fd549a05d5f8124a63f78221aa0ec16",
        }

        if ["5", "0", "5"] <= fsl.Info.version().split(".") < ["6", "0", "5"]:
            expected_checksum = fsl_atlas_checksums["old"]
        else:
            expected_checksum = fsl_atlas_checksums["new"]

        if _sha256(atlas_labels) != expected_checksum:
            raise IOError(
                f"{atlas_labels} has an SHA256 checksum ({_sha256(atlas_labels)}) "
                f"differing from expected ({expected_checksum}), "
                f"file may be corrupted and changed with newer version of FSL."
            )
        return atlas_labels

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-JHUDTI81_dseg.tsv",
        )


class JHUTracts01mm(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "JHUTracts0"

    @staticmethod
    def get_atlas_labels():
        import os

        from .check_dependency import check_environment_variable
        from .inputs import _sha256

        fsl_dir = check_environment_variable("FSLDIR", "FSL")
        atlas_labels = os.path.join(
            fsl_dir, "data", "atlases", "JHU", "JHU-ICBM-tracts-maxprob-thr0-1mm.nii.gz"
        )
        expected_checksum = (
            "eb1de9413a46b02d2b5c7b77852097c6f42c8a5d55a5dbdef949c2e63b95354e"
        )
        if _sha256(atlas_labels) != expected_checksum:
            raise IOError(
                f"{atlas_labels} has an SHA256 checksum ({_sha256(atlas_labels)}) "
                f"differing from expected ({expected_checksum}), "
                f"file may be corrupted and changed with newer version of FSL."
            )
        return atlas_labels

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-JHUTract_dseg.tsv",
        )


class JHUTracts251mm(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "JHUTracts25"

    @staticmethod
    def get_atlas_labels():
        import os

        from .check_dependency import check_environment_variable
        from .inputs import _sha256

        fsl_dir = check_environment_variable("FSLDIR", "FSL")
        atlas_labels = os.path.join(
            fsl_dir,
            "data",
            "atlases",
            "JHU",
            "JHU-ICBM-tracts-maxprob-thr25-1mm.nii.gz",
        )
        expected_checksum = (
            "7cd85fa2be1918fc83173e9bc0746031fd4c08d70d6c81b7b9224b5d3da6d8a6"
        )
        if _sha256(atlas_labels) != expected_checksum:
            raise IOError(
                f"{atlas_labels} has an SHA256 checksum ({_sha256(atlas_labels)}) "
                f"differing from expected ({expected_checksum}), "
                f"file may be corrupted and changed with newer version of FSL."
            )
        return atlas_labels

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-JHUTract_dseg.tsv",
        )


class JHUTracts501mm(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "JHUTracts50"

    @staticmethod
    def get_atlas_labels():
        import os

        from .check_dependency import check_environment_variable
        from .inputs import _sha256

        fsl_dir = check_environment_variable("FSLDIR", "FSL")
        atlas_labels = os.path.join(
            fsl_dir,
            "data",
            "atlases",
            "JHU",
            "JHU-ICBM-tracts-maxprob-thr50-1mm.nii.gz",
        )
        expected_checksum = (
            "20ff0216d770686838de26393c0bdac38c8104760631a1a2b5f518bc0bbb470a"
        )
        if _sha256(atlas_labels) != expected_checksum:
            raise IOError(
                f"{atlas_labels} has an SHA256 checksum ({_sha256(atlas_labels)}) "
                f"differing from expected ({expected_checksum}), "
                f"file may be corrupted and changed with newer version of FSL."
            )
        return atlas_labels

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-JHUTract_dseg.tsv",
        )


class AAL2(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "AAL2"

    @staticmethod
    def get_atlas_labels():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-AAL2_dseg.nii.gz",
        )

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-AAL2_dseg.tsv",
        )


class Hammers(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "Hammers"

    @staticmethod
    def get_atlas_labels():
        from clinica.utils.inputs import RemoteFileStructure, get_file_from_server

        hammers_parc = RemoteFileStructure(
            filename="atlas-Hammers_dseg.nii.gz",
            url="https://aramislab.paris.inria.fr/files/software/cat12/CAT12-Atlases/",
            checksum="c034a7bce2dcab390a0b72f4e7d04769eb3fe5b990d0e18d89b0ce73339a5376",
        )
        return get_file_from_server(hammers_parc)

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-Hammers_dseg.tsv",
        )


class LPBA40(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "LPBA40"

    @staticmethod
    def get_atlas_labels():
        from clinica.utils.inputs import RemoteFileStructure, get_file_from_server

        lpba40_parc = RemoteFileStructure(
            filename="atlas-LPBA40_dseg.nii.gz",
            url="https://aramislab.paris.inria.fr/files/software/cat12/CAT12-Atlases/",
            checksum="20826b572bbbdbcdbf28bbd3801dc0c2fed28d1e54bc4fd5027e64ccc6d50374",
        )
        return get_file_from_server(lpba40_parc)

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-LPBA40_dseg.tsv",
        )


class AICHA(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "AICHA"

    @staticmethod
    def get_atlas_labels():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-AICHA_dseg.nii.gz",
        )

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-AICHA_dseg.tsv",
        )


class Neuromorphometrics(AtlasAbstract):
    def __init__(self):
        AtlasAbstract.__init__(self)

    @staticmethod
    def get_name_atlas():
        return "Neuromorphometrics"

    @staticmethod
    def get_atlas_labels():
        from clinica.utils.inputs import RemoteFileStructure, get_file_from_server

        neuromorphometrics_parc = RemoteFileStructure(
            filename="atlas-Neuromorphometrics_dseg.nii.gz",
            url="https://aramislab.paris.inria.fr/files/software/cat12/CAT12-Atlases/",
            checksum="19a50136cd2f8a14357a19ad8a1dc4a2ecb6beb3fc16cb5441f4f2ebaf64a9a5",
        )
        return get_file_from_server(neuromorphometrics_parc)

    @staticmethod
    def get_tsv_roi():
        from os.path import join, realpath, split

        return join(
            split(realpath(__file__))[0],
            "..",
            "resources",
            "atlases",
            "atlas-Neuromorphometrics_dseg.tsv",
        )



"""This module handles stream and log redirection."""

from enum import Enum


class LoggingLevel(str, Enum):
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


def cprint(msg: str, lvl: str = "info") -> None:
    """
    Print message to the console at the desired logging level.

    Args:
        msg (str): Message to print.
        lvl (str): Logging level between "debug", "info", "warning", "error" and "critical".
                   The default value is "info".
    """
    from logging import getLogger

    # Use the package level logger.
    logger = getLogger("clinica")

    # Log message as info level.
    if lvl == LoggingLevel.debug:
        logger.debug(msg=msg)
    elif lvl == LoggingLevel.info:
        logger.info(msg=msg)
    elif lvl == LoggingLevel.warning:
        logger.warning(msg=msg)
    elif lvl == LoggingLevel.error:
        logger.error(msg=msg)
    elif lvl == LoggingLevel.critical:
        logger.critical(msg=msg)
    else:
        pass



import abc
from typing import Callable, Dict, Optional


class Query:
    """Base Query class.

    Examples
    --------
    >>> from clinica.pydra.query import Query
    >>> q = Query()
    >>> len(q)
    0
    """

    _default_queries = {}
    repr_indent = 4

    def __init__(self, query: Optional[Dict] = None):
        self._query = self.format_query(query)

    @property
    def query(self):
        return self._query

    def format_query(self, input_query: Optional[Dict] = None) -> Dict:
        """Format the input query in a form suitable for DataGrabbers.

        If the query is None, return an empty dictionary.

        Otherwise, `format_query` does two things:

            - parse the input query
            - potentially combine the parsed input query with default values

        The input query is a dictionary but has a heterogeneous structure
        depending on whether we wish to query BIDS or CAPS folders.

        More precisely, it can have the following form :

            {"label" : {query dict, possibly incomplete}}

            example : {"T1w": {"suffix": "T1w"}}

        which is the usual format for BIDS queries. In this case, we simply need
        to filter out queries for labels not supported for BIDS (for example we'd
        keep "T1w" but drop "dartel" or "mask_tissues" if these were present since
        these are labels for CAPS queries).

        The input query can also have the following form:

            {"label" : {kwargs for a query maker}}

            example: {"mask_tissues": {'tissue_number': (1, 2), 'modulation': False}}

        which is the usual format for CAPS queries. In this case, we need to
        filter out queries for labels not supported (for example we'd keep "mask_tissues"
        but drop "T1w" or "pet" if these were present since these are labels for BIDS queries).
        And we need to call the appropriate query maker to obtain a proper query dict.
        In this example, the query maker corresponding to the label "mask_tissues" is the
        function `clinica.utils.input_files.t1_volume_native_tpm_in_mni`. We call this
        function with arguments `tissue_number=(1, 2), modulation=False`.

        After parsing, the query structure is the same for both BIDS and CAPS:

            {"label" : {query dict, possibly incomplete} or list of query dicts}

        The query is then completed with potential default values:

            {"label" : {query dict} or list of query dicts}

            example: {"T1w": {"suffix": "T1w", "extension": [".nii.gz"]}}

        Parameters
        ----------
        input_query : Dict, optional
            Raw input query to format.

        Returns
        -------
        formatted_query : Dict
            Resulting formatted query.
        """
        formatted_query = {}
        if not input_query:
            return formatted_query
        for k, q in self.parse_query(input_query).items():
            if isinstance(q, dict):
                formatted_query[k] = {**self.default_query(k), **q}
            elif isinstance(q, list):
                formatted_query[k] = [{**self.default_query(k), **qq} for qq in q]
            else:
                raise TypeError(
                    f"Unexpected type {type(q)} for query {q}."
                    "Only dictionaries and lists are supported."
                )
        return formatted_query

    @abc.abstractmethod
    def parse_query(self, query: Dict) -> Dict:
        pass

    def default_query(self, label: str) -> Dict:
        """Returns the default query corresponding to the given label.
        If there is no such query, returns an empty dict.

        Parameters
        ----------
        label : str
            Label of the default query to look for.

        Returns
        -------
        Dict :
            The default query corresponding to the provided label.
        """
        return self._default_queries.get(label, {})

    def __getitem__(self, name):
        return self._query[name]

    def __iter__(self):
        return iter(self._query)

    def keys(self):
        return self._query.keys()

    def items(self):
        return self._query.items()

    def values(self):
        return self._query.values()

    def __len__(self):
        return len(self._query)

    def __str__(self):
        import json

        return json.dumps(self._query, indent=self.repr_indent)


class BIDSQuery(Query):
    """BIDSQuery class.

    Examples
    --------
    >>> from clinica.pydra.query import BIDSQuery
    >>> q = BIDSQuery({"T1w": {"suffix": "T1w"}})
    >>> len(q)
    1
    >>> q.query
    {'T1w': {'suffix': 'T1w', 'extension': ['.nii.gz']}}
    """

    _default_queries = {
        "T1w": {"suffix": "T1w", "extension": [".nii.gz"]},
        "pet": {"suffix": "pet", "extension": [".nii.gz"]},
    }

    def parse_query(self, query: Dict) -> Dict:
        """Parse the provided query.

        BIDS query are already specified in a proper dictionary format.
        Parsing is therefore equivalent to filtering out queries corresponding
        to invalid labels (for example labels for CAPS).
        """
        return {k: v for k, v in query.items() if k in self._default_queries}


class CAPSQuery(Query):
    """CAPSQuery class."""

    # Query makers are specified in subclasses
    _query_makers = {}

    def parse_query(self, query: Dict) -> Dict:
        """Parse the provided query.

        For CAPS, the user is providing a dictionary of kwargs for some function
        responsible for building the query dictionary (for example
        `clinica.utils.input_files.t1_volume_deformation_to_template`).
        These functions are stored in a mapping and accessed through the `query_maker`
        method which uses the labels to look for these functions.
        Parsing the query is equivalent to calling these functions for each label with
        the associated user-provided kwargs.

        Parameters
        ----------
        query : Dict
            The user provided query to be parsed.

        Returns
        -------
        parsed_query : Dict
            The parsed query.
        """
        parsed_query = {}
        for label, params in query.items():
            query_maker = self._query_maker(label)
            formatted_query = query_maker(**params)
            if len(formatted_query) > 0:
                parsed_query[label] = formatted_query
        return parsed_query

    def _query_maker(self, label: str) -> Callable:
        """Retrieve the query-making-function associated with the
        provided label.

        Parameters
        ----------
        label : str
            The label to use to get the function.

        Returns
        -------
        Callable :
            The function to use to generate the query dictionary for this label.
            If the label does not match any entry, a default maker which return
             an empty dict for any passed parameters is returned.
        """
        return self._query_makers.get(label, lambda **kwargs: {})


class CAPSFileQuery(CAPSQuery):
    """CAPSFileQuery class.

    This class only holds the mapping between labels and query makers
    to be used with the `CAPSFileDataGrabber`.

    Examples
    --------
    >>> from clinica.pydra.query import CAPSFileQuery
    >>> q = CAPSFileQuery({'mask_tissues': {'tissue_number': (1, 2), 'modulation': False}})
    >>> len(q)
    1
    >>> import json
    >>> print(json.dumps(q.query, indent=2))
    {
        "mask_tissues": [
            {
                "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability.nii*",
                "description": "Tissue probability map graymatter based on native MRI in MNI space (Ixi549) without modulation.",
                "needed_pipeline": "t1-volume-tissue-segmentation"
            },
            {
                "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-whitematter_space-Ixi549Space_modulated-off_probability.nii*",
                "description": "Tissue probability map whitematter based on native MRI in MNI space (Ixi549) without modulation.",
                "needed_pipeline": "t1-volume-tissue-segmentation"
            }
        ]
    }
    """

    from clinica.utils.input_files import (
        t1_volume_dartel_input_tissue,
        t1_volume_deformation_to_template,
        t1_volume_native_tpm,
        t1_volume_native_tpm_in_mni,
    )

    _query_makers = {
        "mask_tissues": t1_volume_native_tpm_in_mni,
        "flow_fields": t1_volume_deformation_to_template,
        "pvc_mask_tissues": t1_volume_native_tpm,
        "dartel_input_tissue": t1_volume_dartel_input_tissue,
    }


class CAPSGroupQuery(CAPSQuery):
    """CAPSGroupQuery class.

    This class only holds the mapping between the labels and query makers
    to be used with the `CAPSGroupDataGrabber`.

    Examples
    --------
    >>> from clinica.pydra.query import CAPSGroupQuery
    >>> import json
    >>> q = CAPSGroupQuery({'dartel_template': {'group_label': "UnitTest"}})
    >>> len(q)
    1
    >>> print(json.dumps(q.query, indent=2))
    {
        "dartel_template": {
        "pattern": "group-UnitTest/t1/group-UnitTest_template.nii*",
        "description": "T1w template file of group UnitTest",
        "needed_pipeline": "t1-volume or t1-volume-create-dartel"
        }
    }
    """

    from clinica.utils.input_files import t1_volume_final_group_template

    _query_makers = {
        "dartel_template": t1_volume_final_group_template,
    }


def query_factory(query: Dict, query_type: str) -> Query:
    """Return the Query instance initialized from passed `query`,
    based on provided `query_type`.

    Parameters
    ----------
    query : dict
        The query in dictionary format.

    query_type : {"bids", "caps_file", "caps_group"}
        The type of query to build.

    Returns
    -------
    Query :
        The resulting Query instance.

    Raises
    ------
    ValueError
        If `query_type` is not a valid value.
    """
    if query_type == "bids":
        return BIDSQuery(query)
    elif query_type == "caps_file":
        return CAPSFileQuery(query)
    elif query_type == "caps_group":
        return CAPSGroupQuery(query)
    else:
        raise ValueError(
            f"query_factory() received an invalid argument for 'query_type' : {query_type}"
            f"Supported values are 'bids', 'caps_file', and 'caps_group'."
        )
